<?php
/**
 * TP Applepay Checkout Plugin.
 */
 
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

class WC_Gateway_TPAP2_Plugin extends WC_Payment_Gateway {

	public function __construct() {
 
		$this->id = 'wc_tpapv2';
		$this->icon = '';
		$this->has_fields = false;
		$this->method_title = 'Total Processing Apple Pay';
		$this->method_description = 'Apple Pay button customisable in your payment flow. Shoppers can decide to continue with the normal checkout, or checkout with Apple Pay.';
		$this->supports = array(
			'products',
			'refunds'
		);
		$this->init_form_fields();
		$this->init_settings();
		$this->title = $this->get_option( 'title' );
		$this->description = $this->get_option( 'description' );
		$this->enabled = $this->get_option( 'enabled' );
		$this->iprestrict = $this->get_option( 'iprestrict' );
		$this->restrictedips = $this->get_option( 'restrictedips' );
		$this->onlyadmins = $this->get_option( 'onlyadmins' );
		$this->jsLogging = ($this->get_option( 'jsLogging' ) !== 'no' ? true : false);
		$this->onlyLoggedIn = ($this->get_option( 'onlyLoggedIn' ) !== 'no' ? true : false);
		$this->forceNoShipping = ($this->get_option( 'forceNoShipping' ) !== 'no' ? true : false);
		$this->forceTerms = ($this->get_option( 'forceTerms' ) !== 'no' ? true : false);
		$this->platformBase = $this->get_option( 'platformBase' );
		$this->entityId_test = $this->get_option( 'entityId_test' );
		$this->accessToken_test = $this->get_option( 'accessToken_test' );
		$this->entityId = $this->get_option( 'entityId' );
		$this->accessToken = $this->get_option( 'accessToken' );
		$this->paymentType = $this->get_option( 'paymentType' );
		$this->domainsToVerify = $this->get_option( 'domainsToVerify' );
		//endpoints
		$this->merchantSessionValidationUrl = 'https://tpap2.totalprocessing.com/session';
		$this->domainValidationUrlTP = 'https://applepay-v2.totalprocessing.com/service/registerWcMerchant.php';
		$this->processDecryptionEndpoint = 'https://tpap2.totalprocessing.com/decrypt';
		//apple pay sheet data
		$this->applepayversion = $this->get_option( 'applepayversion' );
		$this->applepaycountryCode = $this->get_option( 'applepaycountryCode' );
		$this->merchantCapabilities = $this->get_option( 'applepaymerchantCapabilities' );
		$this->supportedNetworks = $this->get_option( 'applepaysupportedNetworks' );
		$this->supportedCountries = $this->get_option( 'applepaysupportedCountries' );
		$this->displayName = $this->get_option( 'applepaydisplayName' );
		$this->label = $this->get_option( 'applepaylabel' );
		$this->refreshPage = ($this->get_option( 'refreshPage' ) == 'yes' ? true : false);
		$this->forceCreateAcc = ($this->get_option( 'forceCreateAcc' ) == 'yes' ? true : false);
		$this->shipInclTax = ($this->get_option( 'shipInclTax' ) == 'yes' ? true : false);
		//custom css
		$this->customCss = $this->get_option( 'customCss' );
		//apple pay button styles
		$this->applepayapplepaystyle = $this->get_option( 'applepayapplepaystyle' );
		$this->applepayapplepaytype = $this->get_option( 'applepayapplepaytype' );
		$this->applepayapplepaylanguage = $this->get_option( 'applepayapplepaylanguage' );
		//Button options
		$this->cartButton = $this->get_option( 'cartButton' );
		$this->cartButtonType = $this->get_option( 'cartButtonType' );
		$this->minicartButton = $this->get_option( 'minicartButton' );
		$this->minicartButtonType = $this->get_option( 'minicartButtonType' );
		$this->checkoutButtonPosition = $this->get_option( 'checkoutButtonPosition' );
		$this->checkoutButton = $this->get_option( 'checkoutButton' );
		$this->checkoutButtonType = $this->get_option( 'checkoutButtonType' );
		$this->productButton = $this->get_option( 'productButton' );
		$this->productButtonType = $this->get_option( 'productButtonType' );
		//button locations
		$this->buttonLocations = $this->get_option( 'buttonLocations' );
		$this->miniCartButtonContainer = $this->get_option( 'miniCartButtonContainer' );
		//set file locations
		$this->domainVerificationFileName = 'apple-developer-merchantid-domain-association';
		//set identifiers
		$this->merchantIdentifier = get_option('wc_gateway_tpapv2_merchantIdentifier_v2');
		// Initialise settings
		//Load funcs
		add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );
	}
	
	public function settings_fields($merge=false) {
        $settings=[];
		$settings['general'] = [
            'title'=>'Apple Pay &amp; Total Processing - WooCommerce v2',
            'description'=>'<p>General settings</p>',
            'fields'=> [
    	        'enabled' => [
    				'title' => 'Enable/Disable',
    				'label' => 'Enable Apple Pay',
    				'type' => 'checkbox',
    				'description' => '',
    				'default' => 'no'
    			],
    			'title' => [
    				'title' => 'Title',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'This controls title text during checkout.',
    				'default' => 'ApplePay',
    				'desc_tip' => true,
    			],
    			'description' => [
    				'title' => 'Description',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'This controls the description seen at checkout.',
    				'default' => 'Checkout with Apple Pay',
    				'desc_tip' => true
    			],
    			'iprestrict' => [
    				'title' => 'Button visible to IP list only',
    				'label' => 'Enable IP restriction for Apple Pay?',
    				'type' => 'checkbox',
    				'description' => '',
    				'default' => 'no'
    			],
    			'restrictedips' => [
    				'title' => 'Allowed IPv4 addresses',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'Separate with commas (current IP:'.$_SERVER['REMOTE_ADDR'].')',
    				'default' => '',
    				'desc_tip' => true,
    			],
    			'onlyadmins' => [
    				'title' => 'Button visible to admin logged in users only',
    				'label' => 'Enable admin restriction to Apple Pay?',
    				'type' => 'checkbox',
    				'description' => '',
    				'default' => 'no'
    			],
		        'onlyLoggedIn' => [
    				'title' => 'Logged in users only?',
    				'label' => 'Enable logged in restriction to Apple Pay?',
    				'type' => 'checkbox',
    				'description' => '',
    				'default' => 'no'
    			],
    			'forceNoShipping' => [
    				'title' => 'Enforce no shipping?',
    				'label' => 'Disable shipping details on Apple Pay sheet?',
    				'type' => 'checkbox',
    				'description' => '',
    				'default' => 'no'
    			],
    			'forceTerms' => [
    				'title' => 'Enforce terms checkbox on checkout page',
    				'label' => 'Checkout terms and conditions must be checked before payment sheet',
    				'type' => 'checkbox',
    				'description' => '',
    				'default' => 'no'
    			],
    			'jsLogging' => [
    				'title' => 'Enable console log? <small><em>Default:Off</em></small>',
    				'label' => 'Console.log events',
    				'type' => 'checkbox',
    				'description' => 'Only if requested to be activated by Total Processing and private access is enabled should this be checked.',
    				'default' => 'no',
    				'desc_tip' => true,
    			]
    		]
		];
		$settings['gateways'] = [
		    'title'=>'Apple Pay Gateway selection',
		    'description'=>'<p>Configure the environment to process within below</p>',
		    'fields'=> [
    		    'platformBase' => [
    				'title' => 'Current Endpoint mode',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => 'eu-prod.oppwa.com',
    				'options' => [
    					'eu-prod.oppwa.com' => 'Live',
    					'eu-test.oppwa.com' => 'Test'
    				]
    			],
    			'entityId_test' => [
    				'title' => 'Entity Id (TEST)',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'Enabled channel for Apple Pay.',
    				'default' => '',
    				'desc_tip' => true,
    			],
    			'accessToken_test' => [
    				'title' => 'Access Token (TEST)',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'Provided by Total Processing',
    				'default' => '',
    				'desc_tip' => true,
    			],
    			'entityId' => [
    				'title' => 'Entity Id (LIVE)',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'Enabled channel for Apple Pay.',
    				'default' => '',
    				'desc_tip' => true,
    			],
    			'accessToken' => [
    				'title' => 'Access Token (LIVE)',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'Provided by Total Processing',
    				'default' => '',
    				'desc_tip' => true,
    			],
    			'paymentType' => [
    				'title' => 'Authorisation Type',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'description' => 'Reccomend using DB as (default) direct capture, some aquirers are not compatible with PA.',
    				'default' => "DB",
    				'desc_tip' => true,
    				'options' => [
    					"DB" => "Debit (DB) transaction immediately captures payment",
    				]
    			],
				'domainsToVerify' => [
    				'title' => 'Domains',
    				'type' => 'textarea',
    				'class' => 'textarea',
					'description' => 'Verifaction domains, separate with comma (,)',
    				'desc_tip' => true,
    				'default' => home_url(),
    			],
    		]
		];
		$settings['configuration'] = [
		    'title'=>'Apple Pay payment sheet',
		    'description'=>'<p>When payment sheet is loaded, specific display can be configured</p>',
		    'fields'=> [
    		    'applepaydisplayName' => [
    				'title' => 'Your store name',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'A string of 64 or fewer UTF-8 characters containing the canonical name for your store. The value is displayed in the Touch Bar on supported models of MacBook Pro.',
    				'default' => get_bloginfo('name'),
    				'desc_tip' => true,
    			],
    			'applepaylabel' => [
    				'title' => 'Billing narrative',
    				'type' => 'text',
    				'class' => '',
    				'description' => 'Provide a business name for the label field. Use the same business name people will see when they look for the charge on their bank or credit card statement.',
    				'default' => get_bloginfo('name'),
    				'desc_tip' => true,
    			],
    			'applepaycountryCode' => [
    				'title' => 'Merchant two-letter ISO 3166 country code',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'description' => 'Set this property to the two-letter code for the country where the payment will be processed. Total Processing will determine which country code value to use.',
    				'default' => 'GB',
    				'desc_tip' => true,
    				'options' => $this->arrStaticData('3166')
    			],
    			'applepayversion' => [
    				'title' => 'Apple Pay JS Version#',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'description' => 'Apple Pay JS version number, default is version 5. determines device compatibilty.',
    				'default' => 6,
    				'desc_tip' => true,
    				'options' => [
						0 => 'Auto device detection',
    					3 => 'macOS 10.13, iOS 11.0',
    					4 => 'macOS 10.14.1, iOS 12.1',
    					5 => 'macOS 10.14.2, iOS 12.1.1',
    					6 => 'macOS 10.14.4, iOS 12.2',
    				]
    			],
    			'applepaymerchantCapabilities' => [
    				'title' => 'The payment capabilities supported by the merchant account category.',
    				'type' => 'multiselect',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => [
    					'supports3DS',
    					'supportsDebit',
    					'supportsCredit'
    				],
    				'options' => [
    					"supports3DS" => "3D Secure",
    					"supportsDebit" => "Debit cards",
    					"supportsCredit" => "Credit cards"
    				]
    			],
    			'applepaysupportedNetworks' => [
    				'title' => 'Card schemes enabled',
    				'type' => 'multiselect',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => [
    					'visa',
    					'masterCard'
    				],
    				'options' => [
    					'amex'=> 'American Express',
    					'discover'=> 'Discover',
    					'jcb'=> 'JCB',
    					'masterCard' => 'Mastercard',
    					'visa'=> 'Visa'
    				]
    			],
    			'applepaysupportedCountries' => [
    				'title' => 'Block card issuers from selected countries (if necessary)',
    				'type' => 'multiselect',
    				'class' => 'wc-enhanced-select-nostd',
    				'description' => 'If the requirments of your processor expects blocks on countries allowed to make payments from these countries, supply them here',
    				'default' => [],
    				'desc_tip' => true,
    				'options' => $this->arrStaticData('3166')
    			],
    			'refreshPage' => [
    				'title' => 'Reload action page on dismiss <small><em>Default:On</em></small>',
    				'label' => 'Reload payment page on closure',
    				'type' => 'checkbox',
    				'description' => 'The page the customer is on will be refreshed on close/payment failure',
    				'default' => 'yes',
    				'desc_tip' => true,
    			],
    			'forceCreateAcc' => [
    				'title' => 'Force account creation on order completed',
    				'label' => 'Force account creation?',
    				'type' => 'checkbox',
    				'default' => 'no',
    				'desc_tip' => false,
    			],
    			'shipInclTax' => [
    			    'title' => 'Force shipping methods Tax.incl',
    				'label' => 'Force shipping method tax?',
    				'type' => 'checkbox',
    				'default' => 'no',
    				'desc_tip' => false,
    			],
    			'customCss' => [
    				'title' => 'Custom button css',
    				'type' => 'textarea',
    				'class' => 'textarea',
    				'default' => 'a.applebtn_cart{display: block;margin-bottom: 1rem;height: 2.5rem;}'."\n".'a.applebtn_checkout{display: block;margin-bottom: 1rem;height: 2.5rem;}'."\n".'a.applebtn_mini-cart{}'."\n".'a.applebtn_product{}'."\n",
    			],
			]
	    ];

		$settings['styles'] = [
		    'title'=>'Apple Pay button styles &amp; placement',
		    'description'=>'<p>Select Apple Pay button locations &amp; individual styles.</p><p><em>*To see active preview of button</em>, you must be using an <strong>Apple Pay compatible web browser i.e. Safari</strong>.</p>',
		    'fields'=> [
		        'applepayapplepaystyle' => [
    				'title' => 'Default button colour <br><span style="font-weight:100;"><em>(changes preview)</em></span>',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd changeButtonStyle',
    				'default' => 'black',
    				'options' => [
    					'black' => 'Black (default)',
    					'white' => 'White',
    					'white-outline' => 'White with line'
    				]
    			],
    			'applepayapplepaytype' => [
    				'title' => 'Default button type <br><span style="font-weight:100;"><em>(changes preview)</em></span>',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd changeButtonStyle',
    				'default' => 'plain',
    				'options' => [
    					'plain' => 'Plain (default)',
    					'book' => 'Book',
    					'buy' => 'Buy',
    					'check-out' => 'Check out',
    					'donate' => 'Donate',
    					'subscribe' => 'Subscribe (default for WooSubscription product)'
    				]
    			],
    			'applepayapplepaylanguage' => [
    				'title' => 'Default button language <br><span style="font-weight:100;"><em>(changes preview)</em></span>',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd changeButtonStyle',
    				'default' => 'en-GB',
    				'options' => [
                        '' => 'Auto geo-detection',
                        'en' => 'English (US)',
                        'ar' => 'Arabic',
                        'ca' => 'Catalan',
                        'zh-CN' => 'Chinese (China)',
                        'zh-HK' => 'Chinese (Hong Kong)',
                        'zh-TW' => 'Chinese (Taiwan)',
                        'hr' => 'Croatian',
                        'cs' => 'Czech',
                        'da' => 'Danish',
                        'nl' => 'Dutch',
                        'en-AU' => 'English (Australia)',
                        'en-GB' => 'English (UK)',
                        'fi' => 'Finnish',
                        'fr-CA' => 'French (Canada)',
                        'fr' => 'French (France)',
                        'de' => 'German',
                        'el' => 'Greek',
                        'he' => 'Hebrew',
                        'hi' => 'Hindi',
                        'hu' => 'Hungarian',
                        'id' => 'Indonesian',
                        'it' => 'Italian',
                        'ja' => 'Japanese',
                        'ko' => 'Korean',
                        'ms' => 'Malay',
                        'no' => 'Norwegian',
                        'pl' => 'Polish',
                        'pt' => 'Portuguese (Brazil)',
                        'pt-PT' => 'Portuguese (Portugal)',
                        'ro' => 'Romanian',
                        'ru' => 'Russian',
                        'sk' => 'Slovak',
                        'es-419' => 'Spanish (Latin America)',
                        'es' => 'Spanish (Spain)',
                        'sv' => 'Swedish',
                        'th' => 'Thai',
                        'tr' => 'Turkish',
                        'uk' => 'Ukrainian',
                        'vi' => 'Vietnamese'
    				]
    			],
		        'buttonLocations' => [
    				'title' => 'Display areas',
    				'type' => 'multiselect',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => [
    					'cart',
    					'checkout'
    				],
    				'options' => [
    					'cart'=> 'In view basket/cart page',
    					'checkout'=> 'On checkout page',
    					/*'mini-cart' => 'Inside mini basket/cart',
    					'product' => 'Individual product page'*/
    				]
    			],
    			'cartButton' => [
    				'title' => 'Cart/Basket button colour',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => '',
    				'options' => [
    				    '' => 'Default',
    					'black' => 'Black',
    					'white' => 'White',
    					'white-outline' => 'White with line'
    				]
    			],
    			'cartButtonType' => [
    				'title' => 'Cart/Basket button type',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => 'check-out',
    				'options' => [
    				    '' => 'Default',
    				    'plain' => 'Plain',
    					'book' => 'Book',
    					'buy' => 'Buy',
    					'check-out' => 'Check out',
    					'donate' => 'Donate',
    					'subscribe' => 'Subscribe (default for WooSubscription product)'
    				]
    			],
    			'minicartButton' => [
    				'title' => 'Mini Cart/Basket button colour',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => '',
    				'options' => [
    				    '' => 'Default',
    					'black' => 'Black',
    					'white' => 'White',
    					'white-outline' => 'White with line'
    				]
    			],
    			'minicartButtonType' => [
    				'title' => 'Mini Cart/Basket button type',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => 'check-out',
    				'options' => [
    				    '' => 'Default',
    				    'plain' => 'Plain',
    					'book' => 'Book',
    					'buy' => 'Buy',
    					'check-out' => 'Check out',
    					'donate' => 'Donate',
    					'subscribe' => 'Subscribe (default for WooSubscription product)',
    					'set-up' => 'Set Up'
    				]
    			],
    			'checkoutButtonPosition' => [
    				'title' => 'Checkout button positioning',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => 'lower',
    				'options' => [
    				    'lower' => 'After payment methods',
    					'upper' => 'Before payment form',
    					'both' => 'Both locations'
    				]
    			],
    			'checkoutButton' => [
    				'title' => 'Checkout button colour',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => '',
    				'options' => [
    				    '' => 'Default',
    					'black' => 'Black',
    					'white' => 'White',
    					'white-outline' => 'White with line'
    				]
    			],
    			'checkoutButtonType' => [
    				'title' => 'Checkout button type',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => 'check-out',
    				'options' => [
    				    '' => 'Default',
    				    'plain' => 'Plain',
    					'book' => 'Book',
    					'buy' => 'Buy',
    					'check-out' => 'Check out',
    					'donate' => 'Donate',
    					'subscribe' => 'Subscribe (default for WooSubscription product)'
    				]
    			],
    			'productButton' => [
    				'title' => 'Product button colour',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => '',
    				'options' => [
    				    '' => 'Default',
    					'black' => 'Black',
    					'white' => 'White',
    					'white-outline' => 'White with line'
    				]
    			],
    			'productButtonType' => [
    				'title' => 'Product button type',
    				'type' => 'select',
    				'class' => 'wc-enhanced-select-nostd',
    				'default' => 'buy',
    				'options' => [
    				    '' => 'Default',
    				    'plain' => 'Plain',
    					'book' => 'Book',
    					'buy' => 'Buy',
    					'check-out' => 'Check out',
    					'donate' => 'Donate',
    					'subscribe' => 'Subscribe (default for WooSubscription product)'
    				]
    			],
    			'miniCartButtonContainer' => [
    				'title' => 'Mini cart HTML container',
    				'label' => 'The class name of the element containing the mini cart buttons.',
    				'type' => 'text',
                    'class' => '',
    				'default' => 'woocommerce-mini-cart__buttons',
    			],
	        ],
	        'body'=>''
		];
		$settings['statistics'] = [
		    'title'=>'Apple Pay Order Data',
		    'description'=>'<p>Payment method detailed statistics overview</p>',
		    'body'=> 'reporting1'
		];
		$settings['status'] = [
		    'title'=>'Apple Pay Plugin Status',
		    'description'=>'<p>Debugging or sense check there are no <strong>critical</strong> <span style="color:#a00;"><span class="dashicons dashicons-warning"></span> Error(s)</span> before allowing for public processing.</p><p><strong>Minor</strong> <span style="color:#ffaf20;"><span class="dashicons dashicons-warning"></span> Warning(s)</span> will allow payment processing to continue.</p>',
		    'body'=> 'showStatus'
		];
		if($merge===true){
		    $retArr=[];
		    foreach($settings as $subSettings){
		        if(count($subSettings)>0){
		            if(isset($subSettings['fields'])){
		                if(count($subSettings['fields'])>0){
		                    $retArr = array_merge($retArr,$subSettings['fields']);
		                }
		            }
		        }
		    }
		    return $retArr;
		}
		return $settings;
	}

	public function init_form_fields(){
		$this->form_fields = $this->settings_fields(true);
	}
	
	public function admin_options() {
	?>
	<style type="text/css">
	    table.wc_status_table tbody {font-weight:100!important; font-size: smaller;}
	    table.wc_status_table code {font-weight:100!important; font-size: smaller!important;}
	    table.wc_status_table tbody tr td {vertical-align: middle;}
	    table.wc_status_table tbody tr td button.tp-admin-ajax {font-size:unset;line-height:unset;padding:5px;}
	    td mark.warning {color: #ffaf20;}
	    div#applePayDemo a.applePayBtn {
	        -webkit-appearance:-apple-pay-button;
	        height: 30px;
            width: 150px;
            display:block;
	    }
	    div#applePayDemo a.black {-apple-pay-button-style: black;}
	    div#applePayDemo a.white {-apple-pay-button-style: white;}
	    div#applePayDemo a.white-outline {-apple-pay-button-style: white-outline;}
	    div#applePayDemo a.plain {-apple-pay-button-type: plain;}
	    div#applePayDemo a.book {-apple-pay-button-type: book;}
	    div#applePayDemo a.buy {-apple-pay-button-type: buy;}
	    div#applePayDemo a.check-out {-apple-pay-button-type: check-out;}
	    div#applePayDemo a.donate {-apple-pay-button-type: donate;}
	    div#applePayDemo a.subscribe {-apple-pay-button-type: subscribe;}
	    a.wc_tpapv2_applepayment {height: 30px; width: 150px;}
	</style>
    <div>
		<?php
		    $showSaveBtn=false;
		    $htmlSubMenu='<ul id="settings-sections" class="subsubsub hide-if-no-js">'."\n";
		    $htmlFormData='';
		    $htmlTitle='';
		    $htmlDescription='';
		    $htmlBody='';
            $settingFields = $this->settings_fields(false);
			if(is_array($settingFields)){
                $lastSetting = count($settingFields);
                $qsArray=[];
                wp_parse_str( $_SERVER['QUERY_STRING'], $qsArray );
                if(!isset($qsArray['opt'])){
                    $sectionId='general';
                } else {
                    $sectionId=$qsArray['opt'];
                }
                foreach( $settingFields as $section => $data ) {
                    $lastSetting--;
                    $current='';
                    if(isset($qsArray['opt'])){
                        if($qsArray['opt'] === $section){
                            $current=' current';
                        }
                    } else {
                        if($section==='general'){
                            $current=' current';
                        }
                    }
                    if($section==='general'){
                        $href='';
                    } else {
                        $href='&opt='.$section;
                    }
                    $htmlSubMenu .= '<li><a class="tab'.$current.'" href="' . get_admin_url() . 'admin.php?page=wc-settings&tab=checkout&section='. $this->id . $href . '">' . ucwords($section) . '</a>'; if($lastSetting > 0){$htmlSubMenu .=' | ';} $htmlSubMenu .= '</li>' . "\n";
                    $display='none';
                    if($section===$sectionId){
                        $display='block';
                    }
                    if(isset($data['fields'])){
                        if(count($data['fields'])>0){
                            if($display==='block'){
                                $showSaveBtn=true;
                            }
                            $htmlFormData .= '<table class="form-table" style="display:'.$display.';">'."\n";
                            $htmlFormData .= $this->generate_settings_html( $data['fields'], false );
                            $htmlFormData .= '</table>'."\n";
                        }
                    }
                }
                if(isset($settingFields[$sectionId]['title'])){
                    $htmlTitle=$settingFields[$sectionId]['title'];
                } else {
                    //
                }
                if(isset($settingFields[$sectionId]['description'])){
                    $htmlDescription=$settingFields[$sectionId]['description'];
                } else {
                    //
                }
                if(isset($settingFields[$sectionId]['body'])){
                    $htmlBody=$settingFields[$sectionId]['body'];
                } else {
                    $htmlBody='';
                }
			}
			$htmlSubMenu .= '</ul>' . "\n";
            $htmlSubMenu .= '<br class="clear" />' . "\n";
            if($showSaveBtn===false){
                $htmlSubMenu .= '<style type="text/css">form#mainform p.submit {display:none;}</style>';
            }
		?>
		<?php echo $htmlSubMenu; ?>
		<table style="width:100%;">
		    <tr>
		        <td style="text-align:left; padding-right:2em; width:75%;">
		            <h3><?php echo $htmlTitle; ?></h3>
		            <?php echo $htmlDescription; ?>
                </td>
		        <td style="text-align:right; width:25%;">
		            <img src="<?php echo plugin_dir_url( __DIR__ ).'assets/img/'.'apple-pay-logo@2x.png'; ?>" alt="Apple Pay" style="height:58px; margin-right:24px;">
                    <img src="<?php echo plugin_dir_url( __DIR__ ).'assets/img/'.'tp-logo.png'; ?>" alt="Total Processing" style="height:58px;">
		        </td>
		    </tr>
		    <?php
                if($sectionId==='styles'){
                    echo '<tr><td colspan="2">'. "\n";
                    echo '<div id="applePayDemo">'. "\n";
                    echo '<a lang="'.$this->applepayapplepaylanguage.'" class="applePayBtn '.$this->applepayapplepaystyle.' '.$this->applepayapplepaytype.'"></a>'. "\n";
                    echo '</div>'. "\n";
                    echo '</td></tr>'. "\n";
                    $htmlBody='';
                }
            ?>
        </table>
        <?php if($sectionId==='styles'){ ?>
        <script type="text/javascript">
        jQuery(function(){
            jQuery('body').on('change', '.changeButtonStyle', function(e){
                var el=document.getElementById('applePayDemo');
                var apStyle = document.getElementById('woocommerce_'+'<?php echo $this->id ?>'+'_applepayapplepaystyle').value;
                var apType = document.getElementById('woocommerce_'+'<?php echo $this->id ?>'+'_applepayapplepaytype').value;
                var apLang = document.getElementById('woocommerce_'+'<?php echo $this->id ?>'+'_applepayapplepaylanguage').value;
                el.innerHTML='';
                var aNode = document.createElement('a');
                aNode.classList.add('applePayBtn');
                aNode.classList.add(apStyle);
                aNode.classList.add(apType);
                aNode.setAttribute('lang',apLang);
                el.appendChild(aNode);
            });
        });
        </script>
        <?php } ?>
        <?php echo $htmlFormData; ?>
        <?php
            if($htmlBody==='showStatus') {
                echo '<div id="statusTablesContainer">'."\n";
                echo $this->generateStatusArray(true);
                echo '</div>'."\n";
            } else if($htmlBody==='reporting1'){
                echo '<div id="statusTablesContainer">'."\n";
                //echo $this->generateReportContent();
                echo '</div>'."\n";
            } else {
                echo $htmlBody;
            }
            if($sectionId === 'status') {
        ?>
        <!--js-->
        <script type="text/javascript">
            jQuery(function(){
                jQuery('body').on('click', '.tp-admin-ajax', function(e){
                    e.preventDefault();
					jQuery(this).prop('disabled', true);
					jQuery(this).text('Requesting....');
                    var action = jQuery(this).data('pl_action');
                    return wp.ajax.post(action,{})
                    .then(function(response) {
                        //console.log(response);
                        if(response.hasOwnProperty('responseData')){
							//console.log(response.responseData);	
						}
                        if(response.valid==true){
                            if(response.hasOwnProperty('containerId')){
                                var el=document.getElementById(response.containerId);
                                if(response.hasOwnProperty('html')){
                                    el.innerHTML=response.html;
                                }
                            }
                        }
                    });
                });
            });
        </script>
        <!--/js-->
		<pre style="display:block;">
        <?php
		
        ?>
		</pre>
		<?php
            }
		?>
    </div>
	<br class="clear" />
	<?php
	}
	
	public function get_icon() {
		$icons_str = '';
		return apply_filters( 'woocommerce_gateway_icon', $icons_str, $this->id );
	}
	
	public function check_plugin(){
        if($this->enabled === 'yes'){
            if($this->onlyadmins === 'yes'){
                if( ! current_user_can( 'administrator' ) ) {
                    return false;
                }
            }
            return true;
        }
        return false;
	}
	
	public function getPluginVer($mainFile){
		$idxFile = str_replace('includes',$mainFile,rtrim(dirname( __FILE__ ), '/'));
		$plugin_data = get_plugin_data( $idxFile );
    	if(is_array($plugin_data)){
			if(isset($plugin_data['Version'])){
				return $plugin_data['Version'];
			}
		}
		return '1.0';
	}
    
	public function run( $plugin_basename ){
        //action hooks
        // Added by Mithilesh K to run this when wp loads
        add_action( 'init', array( $this, '_init' ) );
        if(is_user_logged_in() === true){
			if( current_user_can( 'administrator' ) ) {
				add_action( 'wp_ajax_requestValidation', array( $this, 'requestValidation' ) );
				add_action( 'wp_ajax_nopriv_requestValidation', array( $this, 'requestValidation' ) );
				add_action( 'wp_ajax_moveValidationFile', array( $this, 'moveValidationFile' ) );
				add_action( 'wp_ajax_nopriv_moveValidationFile', array( $this, 'moveValidationFile' ) );
			}
		}
        //check plugin
        if($this->check_plugin()){
            add_action( 'wp_enqueue_scripts', array( $this, 'payment_scripts_tpap' ) );
            if(in_array('cart',$this->buttonLocations)){
                add_action( 'woocommerce_proceed_to_checkout', array( $this, 'draw_express_applepay_tpap' ) );
                if($this->jsLogging === true){
                    add_action( 'woocommerce_after_cart_totals', array( $this, 'draw_test_buttons_tpap' ) );
                }
            }
            if(in_array('checkout',$this->buttonLocations)){
                if($this->checkoutButtonPosition == 'lower'){
                    add_action( 'woocommerce_review_order_before_payment', array( $this, 'draw_express_applepay_tpap' ) );
                } else if($this->checkoutButtonPosition == 'upper') {
                    add_action( 'woocommerce_before_checkout_form', array( $this, 'draw_express_applepay_tpap' ) );
                } else if($this->checkoutButtonPosition == 'both'){
                    add_action( 'woocommerce_before_checkout_form', array( $this, 'draw_express_applepay_tpap' ) );
                    add_action( 'woocommerce_review_order_before_payment', array( $this, 'draw_express_applepay_tpap' ) );
                }
                if($this->jsLogging === true){
                    add_action( 'woocommerce_review_order_before_payment', array( $this, 'draw_test_buttons_tpap' ) );
                }
            }
            if(in_array('mini-cart',$this->buttonLocations)){
                add_action( 'woocommerce_after_mini_cart', array( $this, 'tpap_sideBar_button_visible' ) );
            }
            if(in_array('product',$this->buttonLocations)){
                add_action( 'woocommerce_product_meta_start', array( $this, 'draw_express_applepay_tpap' ) );
                if($this->jsLogging === true){
                    add_action( 'woocommerce_after_single_product', array( $this, 'draw_test_buttons_tpap' ) );
                }
            }
            add_action( 'wp_ajax_applePaymentRequest', array( $this, 'applePaymentRequest' ) );
            add_action( 'wp_ajax_nopriv_applePaymentRequest', array( $this, 'applePaymentRequest' ) );
            add_action( 'wp_ajax_applePayMerchantValidationRequest', array( $this, 'applePayMerchantValidationRequest' ) );
            add_action( 'wp_ajax_nopriv_applePayMerchantValidationRequest', array( $this, 'applePayMerchantValidationRequest' ) );
            add_action( 'wp_ajax_genShipping', array( $this, 'genShipping' ) );
            add_action( 'wp_ajax_nopriv_genShipping', array( $this, 'genShipping' ) );
            add_action( 'wp_ajax_sessUpdateChosenShippingMethod', array( $this, 'sessUpdateChosenShippingMethod' ) );
            add_action( 'wp_ajax_nopriv_sessUpdateChosenShippingMethod', array( $this, 'sessUpdateChosenShippingMethod' ) );
            add_action( 'wp_ajax_onPaymentAuthorized', array( $this, 'onPaymentAuthorized' ) );
            add_action( 'wp_ajax_nopriv_onPaymentAuthorized', array( $this, 'onPaymentAuthorized' ) );
        }
        //filter hooks
        add_filter( 'plugin_action_links_' . $plugin_basename, array( $this, 'plugin_action_links' ) );
        add_filter( 'woocommerce_available_payment_gateways', array( $this, 'tpap_remove_method' ), 20, 1);
        if(!function_exists('is_plugin_active')) {
            include_once(ABSPATH . 'wp-admin/includes/plugin.php');
            if(is_plugin_active( 'woocommerce-multilingual/wpml-woocommerce.php' )){
                add_filter( 'wcml_multi_currency_ajax_actions', array( $this, 'tpap_support_ml_ajax' ), 10, 1 );
            }
        }
    }
	
	public function checkVerificationFileHttpUrl(){
		$url_parse = wp_parse_url( home_url() );
		$url = 'https://'.trailingslashit($url_parse['host']).'.well-known/' . $this->domainVerificationFileName;
		$response = wp_remote_get( $url );
		if ( is_array( $response ) && ! is_wp_error( $response ) ) {
			if($response['response']['code'] === 200){
				return true;
			}
		}
		return false;
	}
    
    public function tpap_support_ml_ajax( $ajax_actions ){
        
        $ajax_actions[] = 'applePaymentRequest';
        $ajax_actions[] = 'applePayMerchantValidationRequest';
        $ajax_actions[] = 'genShipping';
        $ajax_actions[] = 'sessUpdateChosenShippingMethod';
        $ajax_actions[] = 'onPaymentAuthorized';
        $ajax_actions[] = 'createPaymentRequestObject';
        $ajax_actions[] = 'generateLineItems';
        $ajax_actions[] = 'identChosenShipping';
        
        return $ajax_actions;
    }
    
    public function plugin_action_links( $links ) {
		$plugin_links = array(
			'<a href="admin.php?page=wc-settings&tab=checkout&section='.$this->id.'">' . 'Settings' . '</a>',
		);
		return array_merge( $plugin_links, $links );
	}
	
	public function createPaymentRequestObject(){
	    $requiredBillingContactFields = ["postalAddress"];
        $requiredShippingContactFields = ["name","email"];
        $supportedCountries = [];
        $supportedOptions = [];
        if(is_array($this->supportedCountries)){
            $supportedOptions = $this->supportedCountries;
        }
        $diffSupportedCountries = array_diff(array_keys ($this->arrStaticData('3166')), $supportedOptions);
        foreach($diffSupportedCountries as $country){
            $supportedCountries[]=$country;
        }
        //get settings
        $wooAddrOption = get_option( 'woocommerce_ship_to_destination' , '' );
        $billingFields = $this->genAddrReq('GB','billing_');
        $shippingFields = $this->genAddrReq('GB','shipping_');
        if(in_array('billing_phone',$billingFields)){
            $requiredShippingContactFields[]="phone";
        }
        if($wooAddrOption !== 'billing_only'){
            if($this->reqShipping() === true){
                $requiredShippingContactFields[]="postalAddress";
            }
        }
        $array = [
            "countryCode" => $this->applepaycountryCode,
            "currencyCode"=>get_woocommerce_currency(),
            "merchantIdentifier"=>$this->merchantIdentifier,
            "merchantCapabilities"=>$this->merchantCapabilities,
            "supportedNetworks"=>$this->supportedNetworks,
            "supportedCountries" =>$supportedCountries,
            "displayName"=>$this->displayName,
            "total"=>[
                "label"=>$this->label,
                "amount"=>(string)WC()->cart->get_total(null),
            ],
            "lineItems"=>$this->generateLineItems(),
            "requiredBillingContactFields"=>$requiredBillingContactFields,
            "requiredShippingContactFields"=>$requiredShippingContactFields,
            "shippingType"=>"shipping",
            "shippingMethods"=>[],
            "applicationData"=>"",
        ];
        return $array;
	}
	
    public function applePaymentRequest(){
        $array = $this->createPaymentRequestObject();
        wp_send_json_success($array);
	}
    
    public function applePayButtonsArray($buildBtn=false,$showStyle=false){
        $buttonData = [
            'cart' => ['lang'=>$this->applepayapplepaylanguage,'style'=>($this->cartButton == '' ? $this->applepayapplepaystyle : $this->cartButton),'type'=>($this->cartButtonType == '' ? $this->applepayapplepaytype : $this->cartButtonType)],
            'checkout' => ['lang'=>$this->applepayapplepaylanguage,'style'=>($this->checkoutButton == '' ? $this->applepayapplepaystyle : $this->checkoutButton),'type'=>($this->checkoutButtonType == '' ? $this->applepayapplepaytype : $this->checkoutButtonType)],
            'mini-cart' => ['lang'=>$this->applepayapplepaylanguage,'style'=>($this->minicartButton == '' ? $this->applepayapplepaystyle : $this->minicartButton),'type'=>($this->minicartButtonType == '' ? $this->applepayapplepaytype : $this->minicartButtonType)],
            'product' => ['lang'=>$this->applepayapplepaylanguage,'style'=>($this->productButton == '' ? $this->applepayapplepaystyle : $this->productButton),'type'=>($this->productButtonType == '' ? $this->applepayapplepaytype : $this->productButtonType)],
        ];
        if($buildBtn !== false){
            if(isset($buttonData[$buildBtn])){
                $button = $buttonData[$buildBtn];
                if($showStyle == true){
                    return '-webkit-appearance: -apple-pay-button; -apple-pay-button-style:'.$button['style'].'; -apple-pay-button-type:'.$button['type'].';';
                }
                if($button['lang'] == ''){
                    return '<a class="'.$this->id.'_applepayment applebtn_'.$buildBtn.'" style="-webkit-appearance: -apple-pay-button; -apple-pay-button-style:'.$button['style'].'; -apple-pay-button-type:'.$button['type'].';" onclick="javascript:doApplePayFunc2()"></a>';
                }
                return '<a class="'.$this->id.'_applepayment applebtn_'.$buildBtn.'" lang="'.$button['lang'].'" style="-webkit-appearance:-apple-pay-button; -apple-pay-button-style:'.$button['style'].'; -apple-pay-button-type:'.$button['type'].';" onclick="javascript:doApplePayFunc2()"></a>';
            }
            if($this->applepayapplepaylanguage == ''){
                return '<a class="'.$this->id.'_applepayment" style="-webkit-appearance: -apple-pay-button; -apple-pay-button-style:'.$this->applepayapplepaystyle.'; -apple-pay-button-type:'.$this->applepayapplepaytype.';" onclick="javascript:doApplePayFunc2()"></a>';
            }
            return '<a class="'.$this->id.'_applepayment" lang="'.$this->applepayapplepaylanguage.'" style="-webkit-appearance:-apple-pay-button; -apple-pay-button-style:'.$this->applepayapplepaystyle.'; -apple-pay-button-type:'.$this->applepayapplepaytype.';" onclick="javascript:doApplePayFunc2()"></a>';
        }
        return $buttonData;
    }
    
    public function checkPageShouldRun(){
        foreach($this->buttonLocations as $pageType){
            if($pageType === 'cart'){
                if(is_cart()){
                    return true;
                }
            }
            if($pageType === 'checkout'){
                if(is_checkout() === true && is_wc_endpoint_url( 'order-pay' ) !== true){
                    return true;
                }
            }
            if($pageType === 'mini-cart'){
                if(is_woocommerce() === true || is_shop() === true || is_front_page() === true){
                    return true;
                }
            }
            if($pageType === 'product'){
                if(is_product() === true){
                    return true;
                }
            }
        }
        return false;
    }
    
    public function tpap_sideBar_button_visible(){
        if($this->checkPageShouldRun() === true){
            echo '<script type="text/javascript">jQuery(function(){ var tpapv2_fn = window["adjMiniCart"]; if(typeof tpapv2_fn === "function"){adjMiniCart()}});</script>'."\n";
        }
	}
	
	public function applepayversionDetectReturn(){
		$apJsVer = (int)$this->applepayversion;
		if($apJsVer === 0){
			$ua = $_SERVER['HTTP_USER_AGENT'];
			$platforms = ['Macintosh','iPhone','iPad','iPod'];
			$located = false;
			$opsys = false;
			foreach($platforms as $platform){
				if(strpos($ua,$platform) !== false){
					$located = $platform;
					break;
				}
			}
			if($located){
				$firstStr = strstr ( $ua , '(' , false );
				if($firstStr){
					$innerStr = strstr ( $firstStr , ')' , true );
					if($innerStr){
						$opsys = $innerStr;
					}
				}
				
			}
			//return ['ua' => $ua, 'located' => $located, 'opsys' => $opsys];
			return 6;
		}
		return $apJsVer;
	}

	public function payment_scripts_tpap() {
	    if($this->checkPageShouldRun() === true){
			$tp_pluginVer = $this->getPluginVer('woocommerce-gateway-totalprocessing-applepay-v2.php');
	        if($this->jsLogging === true){
				wp_register_script( 'tp_applepay', plugin_dir_url( dirname( __FILE__ ) ).'assets/js/tpJsv2_logging.js?v='.$tp_pluginVer , ['jquery','wp-util'] , null );
	        } else {
                wp_register_script( 'tp_applepay', plugin_dir_url( dirname( __FILE__ ) ).'assets/js/tpJsv2.js?v='.$tp_pluginVer , ['jquery','wp-util'] , null );
	        }
            wp_localize_script( 'tp_applepay', 'tpVars', [
                "subTotalAmount" => (string)WC()->cart->get_total(null),
                "cartPage" => wc_get_cart_url(),
                "miniCart" => $this->applePayButtonsArray('mini-cart',true),
                "refreshPage" => $this->refreshPage,
                "isCheckout" => is_checkout(),
                "isCart" => is_cart(),
                "pluginId" => $this->id,
				"pluginVer" => $tp_pluginVer,
                "adminUrl" => get_admin_url().'admin-ajax.php',
                "version" => $this->applepayversionDetectReturn(),
                "merchantIdentifier" => $this->merchantIdentifier,
                "paymentRequest" => $this->createPaymentRequestObject(),
                "miniCartButtonContainer" => $this->miniCartButtonContainer,
                "forceTerms" => $this->forceTerms
            ]);
			wp_enqueue_style( 'tpap_style', plugin_dir_url( dirname( __FILE__ ) ).'assets/css/tpapv2-style.css?v=' . $tp_pluginVer , [] , null );
            wp_add_inline_style( 'tpap_style', $this->customCss );
            wp_enqueue_script('tp_applepay');
	    }
    }
    
    public function draw_express_applepay_tpap() {
        if($this->checkPageShouldRun() === true){
            if(is_cart()){
                $url = wc_get_cart_url();
                $ref_page = 'cart';
            } else if(is_checkout()){
                $url = wc_get_checkout_url();
                $ref_page = 'checkout';
            } else if(is_product()){
                $url = wc_get_cart_url();
                $ref_page = 'product';
            } else {
                $url = wc_get_cart_url();
                $ref_page = 'cart';
            }
            if($this->onlyLoggedIn === true){
                if(is_user_logged_in() === true){
                    echo $this->applePayButtonsArray($ref_page);
                }
            } else {
                echo $this->applePayButtonsArray($ref_page);
            }
        }
    }
    
    public function draw_test_buttons_tpap(){
        if($this->checkPageShouldRun() === true){
            //echo '<pre>'."\n"; print_r($this->applepayversionDetectReturn()); echo '</pre>'."\n";
        }
    }
    
    public function requestValidation(){
		$response = wp_remote_post(
			$this->domainValidationUrlTP , array(
			'method'      => 'POST',
			'timeout'     => 45,
			'redirection' => 5,
			'httpversion' => '1.0',
			'blocking'    => true,
			'headers'     => array(),
			'body'        => array(
				'domainNames' => $this->domainsToVerify,
				'displayName' => $this->displayName,
				'entityId_test' => $this->entityId_test,
				'entityId' => $this->entityId,
				'entityId_test' => $this->entityId_test,
				'accessToken' => $this->accessToken,
				'accessToken_test' => $this->accessToken_test
				)
			)
		);
		if( is_wp_error( $response ) ) {
			$error_message = $response->get_error_message();
			wp_send_json_success( ["externalUrl"=>$url,"valid"=>true,"containerId"=>"statusTablesContainer","html"=>$this->generateStatusArray(true)] );
		} else {
			$responseData = json_decode($response['body']);
			if(isset($responseData->partnerInternalMerchantIdentifier)){
				//clear current option data.
				delete_option('wc_gateway_tpapv2_merchantIdentifier_v2');
				//legacy plugin option removal.
				delete_option('wc_gateway_tpapv2_certValidationExp');
				delete_option('wc_gateway_tpapv2_domainValidationExp');
				delete_option('wc_gateway_tpapv2_merchantIdentifier');
				delete_option('wc_gateway_tpapv2_certKeyPass');
				delete_option('wc_gateway_tpapv2_e2e');
				//set new merchantIdentifier.
				$this->insUpPluginOption('wc_gateway_tpapv2_merchantIdentifier_v2',$responseData->partnerInternalMerchantIdentifier);
			}
			wp_send_json_success( ["externalUrl"=>$url,"valid"=>true,"containerId"=>"statusTablesContainer","html"=>$this->generateStatusArray(true),"responseData"=>$responseData] );
		}
	}
	
	public function moveValidationFile(){
	    $status=false;
        
		$targetDir = $this->createRootFolderProcedure('.well-known');
		
		$tpapFs_docRoot = $_SERVER["DOCUMENT_ROOT"];
		$tpapFs_homePath = get_home_path();
		$tpapFs_pluginFilePath = plugin_dir_path( __FILE__ );
		
        if($targetDir !== false){
			$location = trailingslashit(rtrim($tpapFs_pluginFilePath,'includes/')) . 'appleValidation/' . $this->domainVerificationFileName;
            $destination = trailingslashit($tpapFs_docRoot) . '.well-known/' . $this->domainVerificationFileName;
            $status = $this->moveFileDir($location,$destination);
        }
		wp_send_json_success( ["valid"=>$status,"containerId"=>"statusTablesContainer","html"=>$this->generateStatusArray(true)] );
	}
	
    public function tpap_remove_method( $available_gateways ){
        $gateway_id = $this->id;
        if(is_checkout()) {
            if(isset($available_gateways[$gateway_id])){
                unset($available_gateways[$gateway_id]);
            }
        }
        return $available_gateways;
    }
    
    public function generateStatusArray($retHtml=false,$colSpan=4){
        $array=[];
        if($retHtml){
            $html='';
        }
        if(!is_admin()){
            return $array;
        }
        $array=[
            [
                "group" => "General Settings",
                "data" => [
                    ["name"=>"WooCommerce Apple Pay payments", "status"=> $this->statusCheckIsEnabled($this->enabled,'',false)],
                    ["name"=>"Apple Pay buttons restricted to only IP(s)", "status"=> $this->statusCheckIsEnabled($this->iprestrict,$this->restrictedips,true,'yes','warning')],
                    ["name"=>"Apple Pay buttons restricted to only admins", "status"=> $this->statusCheckIsEnabled($this->onlyadmins,'',true,'yes','warning')],
                    ["name"=>"Apple Pay console.log()", "status"=> $this->statusCheckIsEnabled($this->get_option( 'jsLogging' ),'',true,($this->onlyadmins == 'yes' ? 'warning' : 'yes'),($this->onlyadmins == 'yes' ? 'warning' : 'warning'))],
                    ["name"=>"SSL Protocol", "status"=> $this->usingSSL()],
                    ["name"=>"Total Processing Communication", "status"=> $this->pingResultStoreOptionResult('wc_gateway_tpapv2_pingTP','https://applepay-v2.totalprocessing.com/service/',true,false)],
                ]
            ],
            [
                "group" => "Gateway Configurations",
                "data" => [
                    ["name"=>"Selected Payment Environment", "status"=> ["valid"=>true,"mark"=> ($this->platformBase != 'eu-test.oppwa.com' ? 'yes' : 'warning') ,"description"=>$this->platformBase,"private"=>false]],
                    ["name"=>"TEST Processing: accessToken", "helptip"=>"", "status"=> $this->stringCheck($this->accessToken_test,'/[a-zA-Z0-9]{58}[=]{2}/m',($this->platformBase != 'eu-test.oppwa.com' ? 'warning' : 'error'),'yes')],
                    ["name"=>"TEST Processing: entityId", "helptip"=>"", "status"=> $this->stringCheck($this->entityId_test,'/[a-z0-9]{32}/m',($this->platformBase != 'eu-test.oppwa.com' ? 'warning' : 'error'),'yes')],
                    ["name"=>"LIVE Processing: accessToken", "helptip"=>"", "status"=> $this->stringCheck($this->accessToken,'/[a-zA-Z0-9]{58}[=]{2}/m',($this->platformBase != 'eu-prod.oppwa.com' ? 'warning' : 'error'),'yes')],
                    ["name"=>"LIVE Processing: entityId", "helptip"=>"", "status"=> $this->stringCheck($this->entityId,'/[a-z0-9]{32}/m',($this->platformBase != 'eu-prod.oppwa.com' ? 'warning' : 'error'),'yes')],
                ]
            ],
            [
                "group" => "Apple Domain Validation",
                "data" => [
                    ["name"=>$this->domainVerificationFileName, "helptip"=>"", "status"=> $this->statusDomainVerifactionFile()],
                ]
            ],
            [
                "group" => "Apple Pay Session Requirements",
                "data" => [
					["name"=>"Apple Pay merchantId (Live)", "helptip"=>"", "status"=> $this->pluginOptionExistsData('wc_gateway_tpapv2_merchantIdentifier_v2','error','yes')],
                ]
            ],
            [
                "group" => "Webhook variables (for Total Processing ref.)",
                "data" => [
                    ["name"=>"var:domain", "status"=> ["valid"=>true,"description"=>$_SERVER['HTTP_HOST'],"private"=>true]],
                    ["name"=>"var:wp_path", "status"=> ["valid"=>true,"description"=>site_url(),"private"=>true]],
                    ["name"=>"get_home_path", "status"=> ["valid"=>true,"description"=>get_home_path(),"private"=>true]],
                    ["name"=>"SERVER_SOFTWARE", "status"=> ["valid"=>true,"description"=>$_SERVER['SERVER_SOFTWARE'] . ' PHP v' . phpversion(),"private"=>false]],
                    ["name"=>"SERVER_PROTOCOL", "status"=> ["valid"=>true,"description"=>$_SERVER['SERVER_PROTOCOL'],"private"=>false]],
                ]
            ],
        ];
        if($retHtml){
            foreach($array as $tbl){
                $html.='<table class="wc_status_table widefat" cellspacing="0">'."\n";
                $html.='<thead>'."\n";
                $html.='<tr>'."\n";
                $html.='<th colspan="'.$colSpan.'" data-export-label="'.$tbl['group'].'">'."\n";
                $html.='<h2>'.$tbl['group'].'</h2></th>'."\n";
                $html.='</tr>'."\n";
                $html.='</thead>'."\n";
                $html.='<tbody>'."\n";
                if(isset($tbl['group']) && is_array($tbl['data'])){
                    foreach($tbl['data'] as $tr){
                        $html.='<tr>'."\n";
                        $html.='<td data-export-label="'.$tr['name'].'">'.$tr['name'].'</td>'."\n";
                        $html.='<td class="help">'."\n";
                        if(isset($tr['helptip'])){
                            $html.='<span class="woocommerce-help-tip">'.$tr['helptip'].'</span>'."\n";
                        }
                        $html.='</td>'."\n";
                        $html.='<td>'."\n";
                        if(isset($tr['status']['mark'])){
                            if((bool)$tr['status']['valid'] === true){
                                $dashIcon = $tr['status']['mark'];
                            } else {
                                $dashIcon = 'warning';
                            }
                            $html.='<mark class="'.$tr['status']['mark'].'"><span class="dashicons dashicons-'.$dashIcon.'"></span> '."\n";
                        }
                        if(isset($tr['status']['private']) && (bool)$tr['status']['private']===true){
                            $html.='<code class="private">'."\n";
                        }
                        $html.= $tr['status']['description'];
                        if(isset($tr['status']['private']) && (bool)$tr['status']['private']===true){
                            $html.='</code>'."\n";
                        }
                        if(isset($tr['status']['mark'])){
                            $html.='</mark>'."\n";
                        }
                        $html.='</td>'."\n";
                        $html.='<td style="text-align:right;">'."\n";
                        if(isset($tr['status']['button'])){
                            $html.='<button class="button tp-admin-ajax" data-pl_action="'.$tr['status']['button']['pl_action'].'" value="'.$tr['status']['button']['value'].'">'.$tr['status']['button']['value'].'</button>'."\n";
                        }
                        $html.='</td>'."\n";
                        $html.='</tr>'."\n";
                    }
                }
                $html.='</tbody>'."\n";
                $html.='</table>'."\n";
            }
            return $html;
        }
        return $array;
    }
    
    public function statusCheckIsEnabled($item,$customDescription='',$reverse=false,$errorClass='error',$successClass='yes'){
        $condition='yes';
        $arrayFalse=["valid"=>false,"mark"=>$errorClass,"description"=>"Disabled","private"=>false];
        $arrayTrue=["valid"=>true,"mark"=>$successClass,"description"=>"Enabled","private"=>false];
        if($reverse){
            $arrayFalse["valid"] = true;
        }
        if($item == $condition){
            $array = $arrayTrue;
        } else {
            $array = $arrayFalse;
        }
        if(trim($customDescription) != ''){
            $array['description'] = $array['description'].' ('. $customDescription.')';
        }
        return $array;
    }
    
    public function stringCheck($item,$pattern,$errorClass='error',$successClass='yes'){
        $arrayFalse=["valid"=>false,"mark"=>$errorClass,"description"=>"","private"=>false];
        $arrayTrue=["valid"=>true,"mark"=>$successClass,"description"=>"","private"=>false];
        preg_match_all($pattern, $item, $matches, PREG_SET_ORDER, 0);
        if(count($matches)>0){
            return $arrayTrue;
        }
        if($pattern == '/[a-zA-Z0-9]{58}[=]{2}/m' || $pattern == '/[a-z0-9]{32}/m'){
            if($errorClass == 'warning'){
                $arrayFalse['description'] = "*Not essential when in current processing environment";
            } else {
                $arrayFalse['description'] = "ERROR processing will fail in selected processing environment!";
            }
        }
        return $arrayFalse;
    }
    
    
	
	public function statusDomainVerifactionFile(){
		$fullPath = '/.well-known/' . $this->domainVerificationFileName;
		$arrayFalse=["valid"=>false,"mark"=>"error","description"=>$fullPath,"private"=>true,"button"=>["value"=>"Move File","pl_action"=>"moveValidationFile"]];
        $arrayTrue=["valid"=>true,"mark"=>"yes","description"=>$fullPath,"private"=>false];
		if($this->checkWellKnownFile()){
			return $arrayTrue;
		}
		return $arrayFalse;
	}
    
    public function pluginOptionExistsData($option,$errorClass='error',$successClass='yes'){
        $value=get_option($option,'Option not set (required)');
        $arrayFalse=["valid"=>false,"mark"=>$errorClass,"description"=>$value.' '.$option,"private"=>true];
        $arrayTrue=["valid"=>true,"mark"=>$successClass,"description"=>$value,"private"=>false];
        if($value !== 'Option not set (required)'){
            if(strlen($value) >= 1){
				if($this->checkWellKnownFile()){
					$arrayTrue["button"]=["value"=>"Re-Initiate Request","pl_action"=>"requestValidation"];
				}
                return $arrayTrue;
            }
        }
		if($this->checkWellKnownFile()){
        	$arrayFalse["button"]=["value"=>"Initiate Request","pl_action"=>"requestValidation"];
		}
        return $arrayFalse;
    }
    
    public function usingSSL($errorClass='error',$successClass='yes'){
        $arrayFalse=["valid"=>false,"mark"=>$errorClass,"description"=>"","private"=>false];
        $arrayTrue=["valid"=>true,"mark"=>$successClass,"description"=>"","private"=>false];
        if(is_ssl()){
            return $arrayTrue;
        }
        return $arrayFalse;
    }
    
    public function pingResultStoreOptionResult($option,$url,$check=false,$retest=false){
        $value=get_option($option,false);
        if($value === false || (int)$value !== 200 || $retest === true){
            $wp_ping=wp_remote_head($url);
            if(!is_wp_error($wp_ping)){
                if(is_array($wp_ping)){
                    if(isset($wp_ping['response'])){
                        if(isset($wp_ping['response']['code'])){
                            $value=(int)$wp_ping['response']['code'];
                            $message=(string)$wp_ping['response']['message'];
                            $url.= ' :'. $value.' ('.$message.')';
                            if(get_option($option) !== false){
                                update_option($option,$value);
                            } else {
                                add_option($option,$value);
                            }
                        }
                    }
                }
            }
        }
        $arrayFalse=["valid"=>false,"mark"=>"error","description"=>$url,"private"=>true, "button"=>["value"=>"Retest","pl_action"=>"pingResultStoreOptionResult"]];
        $arrayTrue=["valid"=>true,"mark"=>"yes","description"=>$url,"private"=>false];
        if((int)$value === 200){
            if($check===true){
                return $arrayTrue;
            }
            return true;
        }
        if($check===true){
            return $arrayFalse;
        }
        return false;
    }
    
    private function createRootFolderProcedure($dir){
        $docBase = $_SERVER['DOCUMENT_ROOT'];
        
        $docPath = trailingslashit($docBase) . $dir;
		
        if(!is_dir($docPath)){
            if(mkdir($docPath)) {
                if(chmod($docPath, 0755)){
                    return $docPath;
                }
            }
        } else {
            return $docPath;
        }
        return false;
    }
    
    private function moveFileDir($location,$destination){
        if(file_exists($location)){
            return rename($location,$destination);
        }
        return false;
    }
    
	public function checkWellKnownFile(){
		if($this->checkVerificationFileHttpUrl()){
			return true;
		}
		$docBase = $_SERVER['DOCUMENT_ROOT'];
        $docPath = trailingslashit($docBase) . '.well-known/' . $this->domainVerificationFileName;
		if(file_exists($docPath)) {
			return true;
		}
		return false;
	}

    public function insUpPluginOption($option,$value,$unset=false){
        $value=sanitize_text_field($value);
        if(get_option($option) !== $unset){
            return update_option($option,$value);
        } else {
            return add_option($option,$value);
        }
    }
    
    
	
	//applePay sheet comms...
	
	public function applePayMerchantValidationRequest(){
        $response = '{"curlError":true}';
		if(isset($_REQUEST["validationUrl"]) && !empty($_REQUEST["validationUrl"])){
            $validation_url = sanitize_text_field($_REQUEST["validationUrl"]);
           	$ch = curl_init();
            curl_setopt($ch, CURLOPT_URL, $this->merchantSessionValidationUrl.'?validationUrl='.$validation_url);
   			curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    		curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
            curl_setopt($ch, CURLOPT_POST, 1);
            curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode(['merchantIdentifier' => $this->merchantIdentifier,'displayName' => $this->displayName, 'initiativeContext' => $_SERVER['HTTP_HOST'], 'initiative' => 'web']));
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);    
			$responseData = curl_exec($ch);
			if(curl_errno($ch)) {
				$responseData = json_encode(['curl_error' => curl_error($ch)]);
			}
			curl_close($ch);
			$response = rtrim($responseData, '1');
        }
        wp_send_json(json_decode($response,true));
	}
	
	public function generateLineItems(){
	    $array=[];
	    //set cartitem line
	    $array[]=['type'=>'final', 'label' => 'Items', 'amount' => number_format((WC()->cart->get_subtotal() + WC()->cart->get_subtotal_tax()), 2)];
	    //set shipping line
	    if($this->reqShipping() === true){
	        $shippingTax = WC()->cart->get_shipping_tax();
	        $shippingTotal = WC()->cart->get_shipping_total();
	        $array[]=['type'=>'final', 'label' => 'Delivery', 'amount' => number_format( ($shippingTotal + $shippingTax) , 2)];
	    }
	    //set discount line
	    if((WC()->cart->get_discount_total() + WC()->cart->get_discount_tax()) > 0){
	        $array[]=['type'=>'final', 'label' => 'Discounts', 'amount' => number_format((WC()->cart->get_discount_total() + WC()->cart->get_discount_tax()), 2)];
	    }
	    return $array;
	}
	
	public function identChosenShipping() {
        $chosen_methods = WC()->session->get( 'chosen_shipping_methods' );
        $chosen_shipping = $chosen_methods[0];
        return $chosen_shipping;
    }
    public function sessUpdateChosenShippingMethod(){
        $id = sanitize_text_field($_POST["id"]);
        WC()->session->set('chosen_shipping_methods', [$id]);
        $chosen_methods = WC()->session->get( 'chosen_shipping_methods' );
        WC()->cart->calculate_totals();
        WC()->cart->calculate_shipping();
        $cartArray = [];
        $cartArray['newTotal'] = ['type'=>'final', 'label' => $this->label, 'amount' => WC()->cart->get_total(null)];
        $cartArray['newLineItems'] = $this->generateLineItems();
        wp_send_json_success( $cartArray );
    }
	public function genShipping() {
        $array = [];
	    if(isset($_POST["countryCode"]) && !empty($_POST["countryCode"])){
            $countryCode = sanitize_text_field($_POST["countryCode"]);
            WC()->customer->set_shipping_country( strtoupper($countryCode) );
	    }
	    if(isset($_POST["postalCode"]) && !empty($_POST["postalCode"])){
            $postalCode = sanitize_text_field($_POST["postalCode"]);
            WC()->customer->set_shipping_postcode( strtoupper($postalCode) );
	    } else {
	        WC()->customer->set_shipping_postcode('');
	    }
	    WC()->cart->calculate_totals();
        WC()->cart->calculate_shipping();
        $cartArray = ['status'=>0];
        $chosen_shipping = $this->identChosenShipping();
        $chosenShippingArray = [];
        $alternateShippingArray = [];
        $cartArray['newTotal'] = ['type'=>'final', 'label' => $this->label, 'amount' => WC()->cart->get_total(null)];
        $cartArray['newLineItems'] = $this->generateLineItems();
        $cartArray['shippingAmount'] = number_format( (WC()->cart->get_shipping_total() + WC()->cart->get_shipping_tax()) , 2);
        $cartArray['shippingMethods'] = [];
        $cartArray['chosenShipping'] = $chosen_shipping;
        $packages = WC()->shipping->get_packages();
        foreach ( $packages as $key => $package ) {
            foreach($package['rates'] as $rate_id => $rate ){
                $runningRateCost=0;
                foreach($rate->taxes as $rateTax){
                    $runningRateCost = $runningRateCost + $rateTax;
                }
                $shipInclTax = number_format( ($rate->cost + $runningRateCost) , 2);
                if($this->shipInclTax === true){
                    $rate->cost = (float)$shipInclTax;
                }
                $cartArray['rates'][$rate->id] = [
                    'id' => $rate->id,
                    'instance_id' => $rate->instance_id,
                    'method_id' => $rate->method_id,
                    'label' => $rate->label,
                    'cost' => $rate->cost
                ];
                if($chosen_shipping === (string)$rate->id){
                    $chosenShippingArray[] = [
                        'label' => $rate->label,
                        'amount' => $rate->cost,
                        'identifier' => $rate->id,
                        'detail' => $rate->label
                    ];
                } else {
                    $alternateShippingArray[] = [
                        'label' => $rate->label,
                        'amount' => $rate->cost,
                        'identifier' => $rate->id,
                        'detail' => $rate->label
                    ];
                }
            }
        }
        if(count($chosenShippingArray) === 1){
            foreach($alternateShippingArray as $optionalShippingMethod){
                $chosenShippingArray[] = $optionalShippingMethod;
            }
            $cartArray['shippingMethods'] = $chosenShippingArray;
        } else {
            $cartArray['shippingMethods'] = $alternateShippingArray;
        }
        if(count($cartArray['shippingMethods'])===0){
            $cartArray['status'] = 1;
            $cartArray['errors'][] = ['code'=>'shippingContactInvalid','contactField'=>'countryCode','message'=>'No shipping to this address'];
        }
        if(count($cartArray) > 0){
            $array = $cartArray;
        }
        wp_send_json_success( $array );
	}
	
	public function reqShipping(){
	    if($this->forceNoShipping === true){
	        return false;
	    }
        $cart_contents = WC()->cart->cart_contents;
        $needs_shipping = false;
        if ( ! empty( $cart_contents ) ) {
            foreach ( $cart_contents as $cart_item_key => $values ) {
                if ( $values['data']->needs_shipping() ) {
                    $needs_shipping = true;
                    break;
                }
            }
        }
        return $needs_shipping;
	}
	
	public function processTokenTransaction($paymentData){
		$body = wp_json_encode( $paymentData );
        $options = [
            'body'=> $body,
            'headers'=> ['Content-Type'=>'application/json','Authorization' => 'Bearer '.($platformBase === 'eu-prod.oppwa.com' ? $this->accessToken : $this->accessToken_test)],
            'timeout'=> 60,
            'redirection'=> 5,
            'blocking'=> true,
            'httpversion'=>'1.0',
            'sslverify'=>false,
            'data_format'=>'body'
        ];
        $response=wp_remote_post( $this->processDecryptionEndpoint, $options );
	    return json_decode($response['body']);
	}
	
	public function prepareRemoteRequest($url,$payload=false,$customRequest='POST'){
		
		$headers = ['Content-Type' => 'application/x-www-form-urlencoded; charset=UTF-8','Authorization' => 'Bearer '.($this->platformBase === 'eu-prod.oppwa.com' ? $this->accessToken : $this->accessToken_test)];

		$array = [
			'method' => $customRequest,
			'timeout' => 45,
			'redirection' => 5,
			'httpversion' => '1.0',
			'blocking' => true,
			'headers' => $headers
		];
		if($payload !== false){
		    $array['body']=$payload;
		}
		
		//$this->writeLog($url,'debug');
		
		$response = wp_remote_request( $url , $array);
		
		//$this->shouldLogResponseHandler($response, $array);
		
		$responseData = json_decode($response['body']);
		
		return $responseData;
	}
	
	public function parseResponseData($responseData){
        global $woocommerce;
        $array = ['status' => false, 'notices'=>[]];
        if(isset($responseData->id)){
            $paymentSuccess=false;
            $transaction_id = (string)$responseData->id;
            if(isset($responseData->result->code)){
                $code = (string)$responseData->result->code;
                if($code == '000.000.000' || $code == '000.100.110'){
                    $paymentSuccess=true;
                }
            }
            if(isset($responseData->result->description)){
                $description = str_replace("'", "" , $responseData->result->description);
            }
            if(isset($responseData->merchantTransactionId)){
                $checks = true;
                $order_id = (int)$responseData->merchantTransactionId;
                $order = wc_get_order( $order_id );
                $amount = number_format($order->get_total(), 2, '.', '');
                $order_data = $order->get_data();
                if($order_data['payment_method'] !== $this->id){
                    $checks = false;
                    //$array['notices'][] = 'payment_method';
                }
                if($order_data['status'] !== 'pending'){
                    $checks = false;
                    //$array['notices'][] = 'status!pending';
                }
                if($amount !== number_format($responseData->customParameters->SHOPPER_amount,2,'.','')){
                    $checks = false;
                    //$array['notices'][] = 'amount mismatch';
                }
                if($order_data['currency'] !== (string)$responseData->customParameters->SHOPPER_currency){
                    $checks = false;
                    //$array['notices'][] = 'currency mismatch';
                }
                if($order_data['order_key'] !== (string)$responseData->customParameters->SHOPPER_order_key){
                    $checks = false;
                    //$array['notices'][] = 'order_key mismatch';
                }
                if($order_data['cart_hash'] !== (string)$responseData->customParameters->SHOPPER_cart_hash){
                    $checks = false;
                    $array['notices'][] = 'Cart items/amount have changed, please retry paying for this order.';
                }
                if($checks){
                    if($paymentSuccess === true){
                        //good to go..
                        $order->payment_complete();
                        //if PA// capture
                        //deprecated
                        //$order->reduce_order_stock();
                        //new
                        wc_reduce_stock_levels($order_id);
                        $order->add_order_note( 'Your order is paid! Thank you!', true );
                        $order->add_order_note( $code, false );
                        $order->set_transaction_id( $transaction_id );
                        $order->add_meta_data('platformBase', $this->platformBase);
                        $order->add_meta_data('paymentType', (string)$responseData->paymentType);
                        $order->save();
                        $woocommerce->cart->empty_cart();
                        $array['status'] = true;
                        $array['redirect'] = $this->get_return_url( $order );
                    } else {
                        $order->add_order_note( $code, false );
                        $order->save();
                        //decline reason..
                        $array['notices'][] = 'Payment not completed: '.$description;
                        $array['error'] = $description;
                    }
                }
            } else {
                $array['notices'][] = 'No order_id found, please retry payment.';
                //perform reversal on $paymentSuccess && $responseData->id
            }
        } else {
            $array['notices'][] = 'No transaction id found, please retry payment.';
        }
        wc_clear_notices();
        foreach($array['notices'] as $notice){
            wc_add_notice( __($notice), 'error');
        }
        return $array;
    }
    
    public function unsetEmptyArrayVars($array){
        foreach($array as $k => $v){
	        if(is_array($v)){
	            foreach($v as $k1 => $v1){
	                if(empty($v1)){
	                    unset($array[$k][$k1]);
	                }
	            }
	            if(count($v)===0){
	                unset($array[$k]);
	            }
	        } else {
	            if(empty($v)){
	                unset($array[$k]);
	            }
	        }
	    }
        return $array;
    }
    
    protected function setCustomerAddressField( $field, $key, $data ) {
        $billing_value  = null;
        $shipping_value = null;
        if ( isset( $data[ "billing_{$field}" ] ) && is_callable( array( WC()->customer, "set_billing_{$field}" ) ) ) {
            $billing_value  = $data[ "billing_{$field}" ];
        }
        if ( isset( $data[ "shipping_{$field}" ] ) && is_callable( array( WC()->customer, "set_shipping_{$field}" ) ) ) {
            $shipping_value = $data[ "shipping_{$field}" ];
        }
        if ( ! is_null( $billing_value ) && is_callable( array( WC()->customer, "set_billing_{$field}" ) ) ) {
            WC()->customer->{"set_billing_{$field}"}( $billing_value );
        }
        if ( ! is_null( $shipping_value ) && is_callable( array( WC()->customer, "set_shipping_{$field}" ) ) ) {
            WC()->customer->{"set_shipping_{$field}"}( $shipping_value );
        }
    }
    
    protected function storeSessAddressVars($data){
        $address_fields=$this->arrStaticData('address_fields');
        //process new data
        array_walk( $address_fields, array( $this, 'setCustomerAddressField' ), $data );
        //save customer data
        WC()->customer->save();
    }
	
	protected function clearSessAddressVars(){
	    //clear sess data
	    $address_fields=$this->arrStaticData('address_fields');
        foreach($address_fields as $k => $field){
            if ( is_callable( array( WC()->customer, "set_billing_{$field}" ) ) ) {
                WC()->customer->{"set_billing_{$field}"}("");
            }
            if ( is_callable( array( WC()->customer, "set_shipping_{$field}" ) ) ) {
                WC()->customer->{"set_shipping_{$field}"}("");
            }
        }
        WC()->customer->save();
	}
	
	public function prepareCustomerIdForOrder( $dataArray ){
        $tpap_customer_id = wc_create_new_customer( $dataArray['billing_email'] , $dataArray['billing_email'] , wp_generate_password( 12, false ) );
        if((int)$tpap_customer_id > 0){
            if(isset($dataArray['billing_first_name']) && !empty($dataArray['billing_first_name'])){
                update_user_meta( $tpap_customer_id, "billing_first_name", $dataArray['billing_first_name'] );
                update_user_meta( $tpap_customer_id, "first_name", $dataArray['billing_first_name'] );
            }
            if(isset($dataArray['billing_last_name']) && !empty($dataArray['billing_last_name'])){
                update_user_meta( $tpap_customer_id, "billing_last_name", $dataArray['billing_last_name'] );
                update_user_meta( $tpap_customer_id, "last_name", $dataArray['billing_last_name'] );
            }
            if(isset($dataArray['billing_address_1']) && !empty($dataArray['billing_address_1'])){
                update_user_meta( $tpap_customer_id, "billing_address_1", $dataArray['billing_address_1'] );
            }
            if(isset($dataArray['billing_city']) && !empty($dataArray['billing_city'])){
                update_user_meta( $tpap_customer_id, "billing_city", $dataArray['billing_city'] );
            }
            if(isset($dataArray['billing_state']) && !empty($dataArray['billing_state'])){
                update_user_meta( $tpap_customer_id, "billing_state", $dataArray['billing_state'] );
            }
            if(isset($dataArray['billing_postcode']) && !empty($dataArray['billing_postcode'])){
                update_user_meta( $tpap_customer_id, "billing_postcode", $dataArray['billing_postcode'] );
            }
            if(isset($dataArray['billing_country']) && !empty($dataArray['billing_country'])){
                update_user_meta( $tpap_customer_id, "billing_country", $dataArray['billing_country'] );
            }
            if(isset($dataArray['billing_email']) && !empty($dataArray['billing_email'])){
                update_user_meta( $tpap_customer_id, "billing_email", $dataArray['billing_email'] );
            }
            if(isset($dataArray['billing_phone']) && !empty($dataArray['billing_phone'])){
                update_user_meta( $tpap_customer_id, "billing_phone", $dataArray['billing_phone'] );
            }
            return $tpap_customer_id;
        }
        return 0;
	}
	
	public function onPaymentAuthorized(){
	    global $woocommerce;
	    $jsonData=file_get_contents('php://input');
	    if($this->jsLogging === true){
	        $this->writeLog($jsonData);
	    }
	    $postData=json_decode($jsonData,true);
	    $readyState=false;
	    $errors=[];
	    $chkArray=[];
	    $fieldMapping = [
	        'billing_email' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'emailAddress','message'=>'Email '], 'func'=>'validateEmail' ],
	        'billing_phone' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'phoneNumber','message'=>'Phone '], 'func'=>'validatePhone' ],
	        'billing_first_name' => [ 'error'=> ['code'=>'billingContactInvalid','contactField'=>'name','message'=>'Name '] ],
	        'billing_last_name' => [ 'error'=> ['code'=>'billingContactInvalid','contactField'=>'name','message'=>'Name '] ],
	        'billing_address_1' => [ 'error'=> ['code'=>'billingContactInvalid','contactField'=>'addressLines','message'=>'Address '] ],
	        'billing_city' => [ 'error'=> ['code'=>'billingContactInvalid','contactField'=>'locality','message'=>'Town/City '] ],
	        'billing_state' => [ 'error'=> ['code'=>'billingContactInvalid','contactField'=>'administrativeArea','message'=>'State/County '] ],
	        'billing_postcode' => [ 'error'=> ['code'=>'billingContactInvalid','contactField'=>'postalCode','message'=>'Postal/Zip '], 'func'=>'validatePostcodeBilling' ],
	        'billing_country' => [ 'error'=> ['code'=>'billingContactInvalid','contactField'=>'countryCode','message'=>'Country '] ],
	        'shipping_first_name' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'name','message'=>'Name '] ],
	        'shipping_last_name' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'name','message'=>'Name '] ],
	        'shipping_address_1' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'addressLines','message'=>'Address '] ],
	        'shipping_city' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'locality','message'=>'Town/City '] ],
	        'shipping_state' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'administrativeArea','message'=>'State/County '] ],
	        'shipping_postcode' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'postalCode','message'=>'Postal/Zip '], 'func'=>'validatePostcodeShipping' ],
	        'shipping_country' => [ 'error'=> ['code'=>'shippingContactInvalid','contactField'=>'countryCode','message'=>'Country '] ],
	    ];
	    $billing = $this->unsetEmptyArrayVars($postData["billingContact"]);
	    $shipping = $this->unsetEmptyArrayVars($postData["shippingContact"]);
	    if(!isset($billing["countryCode"])){
	        $errors[]=['code'=>'billingContactInvalid','contactField'=>'countryCode','message'=>'Country is mandatory'];
	    } else {
	        $chkArray = array_merge($chkArray,$this->genAddrReq($billing["countryCode"],'billing_'));
	    }
	    $paymentRequest = $this->createPaymentRequestObject();
	    if(in_array("postalAddress",$paymentRequest["requiredShippingContactFields"])){
	        if(!isset($shipping["countryCode"])){
    	        $errors[]=['code'=>'shippingContactInvalid','contactField'=>'countryCode','message'=>'Country is mandatory'];
    	    } else {
    	        $chkArray = array_merge($chkArray,$this->genAddrReq($shipping["countryCode"],'shipping_'));
    	    }
	    }
	    $dataArray = $this->prepareAuthorisedDataArray($billing,$shipping);
	    foreach($chkArray as $chk){
	        if(!isset($dataArray[$chk])){
	            $error = $fieldMapping[$chk]['error'];
	            $error['message'] = $error['message'] . "required";
	            $errors[]=$error;
	        } else {
	            if(isset($fieldMapping[$chk]['func'])){
	                $func=$fieldMapping[$chk]['func'];
	                if($func==='validateEmail'){
	                    $validate = is_email( sanitize_email($dataArray[$chk]) );
	                } else if($func==='validatePhone'){
	                    $validate = WC_Validation::is_phone( $dataArray[$chk] );
	                } else if($func==='validatePostcodeBilling'){
	                    $validate = WC_Validation::is_postcode( $dataArray[$chk], $dataArray['billing_country'] );
	                } else if($func==='validatePostcodeShipping'){
	                    $validate = WC_Validation::is_postcode( $dataArray[$chk], $dataArray['shipping_country'] );
	                } else {
	                    $validate = true;
	                }
	                if($validate === false){
	                    $error = $fieldMapping[$chk]['error'];
	                    $error['message'] = $error['message'] . "invalid";
	                    $errors[]=$error;
	                }
	            }
	        }
	    }
	    if(count($errors) === 0){
	        $readyState = true;
	        //clear session vars
	        $this->clearSessAddressVars();
	        //set session vars
	        $this->storeSessAddressVars($dataArray);
	    }
        if($readyState === true){
            $dataArray['payment_method']=$this->id;
            $dataArray['customer_id'] = get_current_user_id();
            $checkout = new WC_Checkout();
            if($dataArray['customer_id'] === 0){
                $customer_id = email_exists( $dataArray['billing_email'] );
                if(!is_wp_error($customer_id) && (int)$customer_id > 0){
                    $dataArray['customer_id'] = $customer_id;
                } else if($this->forceCreateAcc === true || $checkout->is_registration_required() === true){
                    $dataArray['customer_id'] = $this->prepareCustomerIdForOrder( $dataArray );
                }
            }
            
            $woocommerce->cart->calculate_totals();
            $order_id = $checkout->create_order( $dataArray );
            if((int)$order_id > 0){
                update_post_meta($order_id, '_customer_user', $dataArray['customer_id']);
                $paymentData = [];
                $paymentData['order_id'] = $order_id;
                $order = wc_get_order( $order_id );
                $amount = number_format($order->get_total(), 2, '.', '');
                $order_data = $order->get_data();
                $payload = [
					"entityId" => ($this->platformBase === 'eu-prod.oppwa.com' ? $this->entityId : $this->entityId_test),
                    "paymentType" => $this->paymentType,
                    "amount" => $amount,
                    "currency" => $order_data['currency'],
					"applePay.paymentToken" => $postData['token'],
                    "merchantTransactionId" => $order_data['id'],
                    "customer.merchantCustomerId" => $order_data['customer_id'],
					"customParameters[SHOPPER_endpoint]" => $this->platformBase,
					"customParameters[SHOPPER_bearer]" => 'Bearer '.($this->platformBase === 'eu-prod.oppwa.com' ? $this->accessToken : $this->accessToken_test),
                    "customParameters[SHOPPER_amount]" => $amount,
                    "customParameters[SHOPPER_currency]" => $order_data['currency'],
                    "customParameters[SHOPPER_order_key]" => $order_data['order_key'],
                    "customParameters[SHOPPER_cart_hash]" => $order_data['cart_hash'],
					"customParameters[SHOPPER_cart_hash]" => $order_data['cart_hash'],
                    "card.holder" => (string)($order_data['billing']['first_name'].' '. $order_data['billing']['last_name']),
                    "customer.givenName" => $order_data['billing']['first_name'],
                    "customer.surname" => $order_data['billing']['last_name'],
                    "customer.email" => $order_data['billing']['email'],
                    "customer.ip" => $order_data['customer_ip_address'],
                    "customer.browserFingerprint.value" => $order_data['customer_user_agent']
                ];
                if(isset($order_data['billing']['phone'])){
                    if(!empty($order_data['billing']['phone'])){
                        $payload["customer.mobile"] = $order_data['billing']['phone'];
                    }
                }
                if(isset($order_data['billing']['address_1'])){
                    if(!empty($order_data['billing']['address_1'])){
                        $payload["billing.street1"] = $order_data['billing']['address_1'];
                    }
                }
                if(isset($order_data['billing']['address_2'])){
                    if(!empty($order_data['billing']['address_2'])){
                        $payload["billing.street2"] = $order_data['billing']['address_2'];
                    }
                }
                if(isset($order_data['billing']['city'])){
                    if(!empty($order_data['billing']['city'])){
                        $payload["billing.city"] = $order_data['billing']['city'];
                    }
                }
                if(isset($order_data['billing']['state'])){
                    if(!empty($order_data['billing']['state'])){
                        $payload["billing.state"] = $order_data['billing']['state'];
                    }
                }
                if(isset($order_data['billing']['postcode'])){
                    if(!empty($order_data['billing']['postcode'])){
                        $payload["billing.postcode"] = $order_data['billing']['postcode'];
                    }
                }
                if(isset($order_data['billing']['country'])){
                    if(!empty($order_data['billing']['country'])){
                        $payload["billing.country"] = $order_data['billing']['country'];
                    }
                }
                if(isset($order_data['shipping']['address_1'])){
                    if(!empty($order_data['shipping']['address_1'])){
                        $payload["shipping.street1"] = $order_data['shipping']['address_1'];
                    }
                }
                if(isset($order_data['shipping']['address_2'])){
                    if(!empty($order_data['shipping']['address_2'])){
                        $payload["shipping.street2"] = $order_data['shipping']['address_2'];
                    }
                }
                if(isset($order_data['shipping']['city'])){
                    if(!empty($order_data['shipping']['city'])){
                        $payload["shipping.city"] = $order_data['shipping']['city'];
                    }
                }
                if(isset($order_data['shipping']['state'])){
                    if(!empty($order_data['shipping']['state'])){
                        $payload["shipping.state"] = $order_data['shipping']['state'];
                    }
                }
                if(isset($order_data['shipping']['postcode'])){
                    if(!empty($order_data['shipping']['postcode'])){
                        $payload["shipping.postcode"] = $order_data['shipping']['postcode'];
                    }
                }
                if(isset($order_data['shipping']['country'])){
                    if(!empty($order_data['shipping']['country'])){
                        $payload["shipping.country"] = $order_data['shipping']['country'];
                    }
                }
                $paymentData['payload'] = $payload;
                $responseData = $this->prepareRemoteRequest($this->processDecryptionEndpoint, $payload);
                $paymentData['decryptionResponse'] = $responseData;
                $parseResult = $this->parseResponseData($responseData);
                $paymentData['parseResult'] = $parseResult;
                if($parseResult['status']===true){
                    wp_send_json_success( ['returnMessage' => [
                        'status'=> 0
                    ],
                    'redirect'=>$parseResult['redirect']
                    ] );
                } else {
                    wp_send_json_success( ['returnMessage' => [
                        'status'=>0,
                    ],
                    'error'=>$parseResult['error']
                    ] );
                }
            }
        }
        wp_send_json_success( ['returnMessage' => [
            'status'=>1,
            'errors'=> [[
                'code'=>'unknown',
                'contactField'=>null,
                'message' => null
            ]]
        ]] );
	}
	
	public function genAddrReq($countryCode = 'GB',$type = 'billing_'){
        $countries = new WC_Countries();
        $addrFields = $countries->get_address_fields($countryCode,$type);
        $reqArray = [];
        foreach($addrFields as $field => $req){
            if($req['required'] == 1){
                $reqArray[] = $field;
                continue;
            }
        }
        return $reqArray;
	}
	
	//end applePay sheet comms
	
	public function payment_fields() {

	}
	
	public function validate_fields(){
		return true;
	}
	
	public function process_payment( $order_id ) {
		return;
	}
	
	public function process_refund( $order_id, $amount = null, $reason = '') {
		global $woocommerce;
		$order = new WC_Order($order_id);
		$order_data = $order->get_data();
		if($amount !== null){
		    $amount = number_format($amount, 2, '.', '');
		} else {
		    $amount = '0.00';
		}
		$id = $order->get_transaction_id();
		$platformBase = $order->get_meta('platformBase');
		$paymentType = $order->get_meta('paymentType');
		if(in_array($paymentType,['DB','CP','RB'])){
		    $payload = ['paymentType'=>'RF','currency'=>$order_data['currency'],'amount'=>$amount];
		} else if(in_array($paymentType,['PA'])){
		    $payload = ['paymentType'=>'RV','currency'=>$order_data['currency']];
		} else {
		    return new WP_Error( 'Error', 'Original paymentType not recognised.' );
		}
		$payload['entityId'] = ($platformBase == 'eu-prod.oppwa.com' ? $this->entityId : $this->entityId_test);
	    $array = [
			'method' => 'POST',
			'timeout' => 45,
			'redirection' => 5,
			'httpversion' => '1.0',
			'blocking' => true,
			'headers' => ['Content-Type' => 'application/x-www-form-urlencoded; charset=UTF-8','Authorization' => 'Bearer '.($platformBase == 'eu-prod.oppwa.com' ? $this->accessToken : $this->accessToken_test)],
			'body' => $payload
		];
		$response = wp_remote_request( 'https://'.$platformBase.'/v1/payments/'.$id , $array);
		$result = json_decode($response['body']);
		if(!is_object($result)){
		    return new WP_Error( 'Error', 'Transaction response error.' );
		}
		if($result->result->code == '000.000.000' || $result->result->code == '000.100.110'){
		    return true;
		} else {
		    return new WP_Error( 'Error', 'Transaction refused: '.$result->result->description );
		}
		return new WP_Error( 'Error', 'Transaction processing error' );
	}
	
	public function prepareAuthorisedDataArray($billing,$shipping){
	    $dataArray = [];
	    if(isset($billing['emailAddress'])){
            if(!empty($billing['emailAddress'])){
                $dataArray['billing_email'] = sanitize_text_field($billing['emailAddress']);
            }
        }
        if(isset($billing['phoneNumber'])){
            if(!empty($billing['phoneNumber'])){
                $dataArray['billing_phone'] = sanitize_text_field($billing['phoneNumber']);
            }
        }
        if(isset($billing['givenName'])){
            if(!empty($billing['givenName'])){
                $dataArray['billing_first_name'] = sanitize_text_field($billing['givenName']);
            }
        }
        if(isset($billing['familyName'])){
            if(!empty($billing['familyName'])){
                $dataArray['billing_last_name'] = sanitize_text_field($billing['familyName']);
            }
        }
        if(isset($billing['countryCode'])){
            if(!empty($billing['countryCode'])){
                $dataArray['billing_country'] = strtoupper(sanitize_text_field($billing['countryCode']));
            }
        }
        if(isset($billing['addressLines'])){
            if(is_array($billing['addressLines'])){
                if(!empty($billing['addressLines'][0])){
                    $dataArray['billing_address_1'] = sanitize_text_field($billing['addressLines'][0]);
                }
                if(isset($billing['addressLines'][1])){
                    if(!empty($billing['addressLines'][1])){
                        $dataArray['billing_address_2'] = sanitize_text_field($billing['addressLines'][1]);
                    }
                }
            }
        }
        if(isset($billing['locality'])){
            if(!empty($billing['locality'])){
                $dataArray['billing_city'] = sanitize_text_field($billing['locality']);
            }
        }
        if(isset($billing['administrativeArea'])){
            if(!empty($billing['administrativeArea'])){
                $dataArray['billing_state'] = sanitize_text_field($billing['administrativeArea']);
            }
        }
        if(isset($billing['postalCode'])){
            if(!empty($billing['postalCode'])){
                $dataArray['billing_postcode'] = sanitize_text_field($billing['postalCode']);
            }
        }
        //check alt subs.
        if(isset($billing['subLocality'])){
            if(!empty($billing['subLocality'])){
                if(isset($dataArray['billing_city'])){
                    $dataArray['billing_city'] .= ' '. sanitize_text_field($billing['subLocality']);
                } else {
                    $dataArray['billing_city'] = sanitize_text_field($billing['subLocality']);
                }
            }
        }
        if(isset($billing['subAdministrativeArea'])){
            if(!empty($billing['subAdministrativeArea'])){
                if(isset($dataArray['billing_state'])){
                    $dataArray['billing_state'] .= ' '. sanitize_text_field($billing['subAdministrativeArea']);
                } else {
                    $dataArray['billing_state'] = sanitize_text_field($billing['subAdministrativeArea']);
                }
            }
        }
        if(isset($shipping['emailAddress'])){
            if(!empty($shipping['emailAddress']) && !isset($dataArray['billing_email'])){
                $dataArray['billing_email'] = sanitize_text_field($shipping['emailAddress']);
            }
        }
        if(isset($shipping['phoneNumber'])){
            if(!empty($shipping['phoneNumber']) && !isset($dataArray['billing_phone'])){
                $dataArray['billing_phone'] = sanitize_text_field($shipping['phoneNumber']);
            }
        }
        if(isset($shipping['givenName'])){
            if(!empty($shipping['givenName'])){
                $dataArray['shipping_first_name'] = sanitize_text_field($shipping['givenName']);
            }
        }
        if(isset($shipping['familyName'])){
            if(!empty($shipping['familyName'])){
                $dataArray['shipping_last_name'] = sanitize_text_field($shipping['familyName']);
            }
        }
        if(isset($shipping['countryCode'])){
            if(!empty($shipping['countryCode'])){
                $dataArray['shipping_country'] = strtoupper(sanitize_text_field($shipping['countryCode']));
            }
        }
        if(isset($shipping['addressLines'])){
            if(is_array($shipping['addressLines'])){
                if(!empty($shipping['addressLines'][0])){
                    $dataArray['shipping_address_1'] = sanitize_text_field($shipping['addressLines'][0]);
                }
                if(isset($shipping['addressLines'][1])){
                    if(!empty($shipping['addressLines'][1])){
                        $dataArray['shipping_address_2'] = sanitize_text_field($shipping['addressLines'][1]);
                    }
                }
            }
        }
        if(isset($shipping['locality'])){
            if(!empty($shipping['locality'])){
                $dataArray['shipping_city'] = sanitize_text_field($shipping['locality']);
            }
        }
        if(isset($shipping['administrativeArea'])){
            if(!empty($shipping['administrativeArea'])){
                $dataArray['shipping_state'] = sanitize_text_field($shipping['administrativeArea']);
            }
        }
        if(isset($shipping['postalCode'])){
            if(!empty($shipping['postalCode'])){
                $dataArray['shipping_postcode'] = sanitize_text_field($shipping['postalCode']);
            }
        }
        //check alt subs.
        if(isset($shipping['subLocality'])){
            if(!empty($shipping['subLocality'])){
                if(isset($dataArray['shipping_city'])){
                    $dataArray['shipping_city'] .= ' '. sanitize_text_field($shipping['subLocality']);
                } else {
                    $dataArray['shipping_city'] = sanitize_text_field($shipping['subLocality']);
                }
            }
        }
        if(isset($shipping['subAdministrativeArea'])){
            if(!empty($shipping['subAdministrativeArea'])){
                if(isset($dataArray['shipping_state'])){
                    $dataArray['shipping_state'] .= ' '. sanitize_text_field($shipping['subAdministrativeArea']);
                } else {
                    $dataArray['shipping_state'] = sanitize_text_field($shipping['subAdministrativeArea']);
                }
            }
        }
        return $dataArray;
	}
    
    // Added by Mithilesh K
    // This runs when wp hooks is triggered
    public function _init(){
        $this->update_platformbase_init();
    }
    
    // Added by Mithilesh K
    // this updates platformBase to latest endpoint to avoid deprecated enpoint
    public function update_platformbase_init(){
        $oldPlatformBase     = $this->get_option( 'platformBase' );
        if( $oldPlatformBase == 'test.oppwa.com' ){
            $this->update_option( 'platformBase', 'eu-test.oppwa.com' );
        }
        if( $oldPlatformBase == 'oppwa.com' ){
            $this->update_option( 'platformBase', 'eu-prod.oppwa.com' );
        }
    }
	
	protected function writeLog($obj){	
	  /*if(is_object($obj) === true || is_array($obj) === true){
	      $obj = json_encode($obj);
	  }
      $filename = $this->logPath.'/'.date('Ymd').'.log';
      $fh = fopen($filename, "a+");
      if(!$fh){
          return false;
      }
      fwrite($fh, date('H:i:s').' => '.$obj."\n");
      fclose($fh);*/
      return true;
    }

	public function arrStaticData($key){
        $array=[
            'address_fields' => [
                'first_name',
                'last_name',
                'company',
                'email',
                'phone',
                'address_1',
                'address_2',
                'city',
                'postcode',
                'state',
                'country'
            ],
            '3166'=>[
                'AD'=>'Andorra',
                'AE'=>'United Arab Emirates',
                'AF'=>'Afghanistan',
                'AG'=>'Antigua and Barbuda',
                'AI'=>'Anguilla',
                'AL'=>'Albania',
                'AM'=>'Armenia',
                'AO'=>'Angola',
                'AQ'=>'Antarctica',
                'AR'=>'Argentina',
                'AS'=>'American Samoa',
                'AT'=>'Austria',
                'AU'=>'Australia',
                'AW'=>'Aruba',
                'AX'=>'Aland Islands',
                'AZ'=>'Azerbaijan',
                'BA'=>'Bosnia and Herzegovina',
                'BB'=>'Barbados',
                'BD'=>'Bangladesh',
                'BE'=>'Belgium',
                'BF'=>'Burkina Faso',
                'BG'=>'Bulgaria',
                'BH'=>'Bahrain',
                'BI'=>'Burundi',
                'BJ'=>'Benin',
                'BL'=>'Saint Barthelemy',
                'BM'=>'Bermuda',
                'BN'=>'Brunei Darussalam',
                'BO'=>'Bolivia (Plurinational State of)',
                'BQ'=>'Bonaire Sint Eustatius and Saba',
                'BR'=>'Brazil',
                'BS'=>'Bahamas',
                'BT'=>'Bhutan',
                'BV'=>'Bouvet Island',
                'BW'=>'Botswana',
                'BY'=>'Belarus',
                'BZ'=>'Belize',
                'CA'=>'Canada',
                'CC'=>'Cocos (Keeling) Islands',
                'CD'=>'Congo Democratic Republic of the',
                'CF'=>'Central African Republic',
                'CG'=>'Congo',
                'CH'=>'Switzerland',
                'CI'=>'Cote dIvoire',
                'CK'=>'Cook Islands',
                'CL'=>'Chile',
                'CM'=>'Cameroon',
                'CN'=>'China',
                'CO'=>'Colombia',
                'CR'=>'Costa Rica',
                'CU'=>'Cuba',
                'CV'=>'Cabo Verde',
                'CW'=>'Curacao',
                'CX'=>'Christmas Island',
                'CY'=>'Cyprus',
                'CZ'=>'Czechia',
                'DE'=>'Germany',
                'DJ'=>'Djibouti',
                'DK'=>'Denmark',
                'DM'=>'Dominica',
                'DO'=>'Dominican Republic',
                'DZ'=>'Algeria',
                'EC'=>'Ecuador',
                'EE'=>'Estonia',
                'EG'=>'Egypt',
                'EH'=>'Western Sahara',
                'ER'=>'Eritrea',
                'ES'=>'Spain',
                'ET'=>'Ethiopia',
                'FI'=>'Finland',
                'FJ'=>'Fiji',
                'FK'=>'Falkland Islands (Malvinas)',
                'FM'=>'Micronesia (Federated States of)',
                'FO'=>'Faroe Islands',
                'FR'=>'France',
                'GA'=>'Gabon',
                'GB'=>'United Kingdom of Great Britain and Northern Ireland',
                'GD'=>'Grenada',
                'GE'=>'Georgia',
                'GF'=>'French Guiana',
                'GG'=>'Guernsey',
                'GH'=>'Ghana',
                'GI'=>'Gibraltar',
                'GL'=>'Greenland',
                'GM'=>'Gambia',
                'GN'=>'Guinea',
                'GP'=>'Guadeloupe',
                'GQ'=>'Equatorial Guinea',
                'GR'=>'Greece',
                'GS'=>'South Georgia and the South Sandwich Islands',
                'GT'=>'Guatemala',
                'GU'=>'Guam',
                'GW'=>'Guinea-Bissau',
                'GY'=>'Guyana',
                'HK'=>'Hong Kong',
                'HM'=>'Heard Island and McDonald Islands',
                'HN'=>'Honduras',
                'HR'=>'Croatia',
                'HT'=>'Haiti',
                'HU'=>'Hungary',
                'ID'=>'Indonesia',
                'IE'=>'Ireland',
                'IL'=>'Israel',
                'IM'=>'Isle of Man',
                'IN'=>'India',
                'IO'=>'British Indian Ocean Territory',
                'IQ'=>'Iraq',
                'IR'=>'Iran (Islamic Republic of)',
                'IS'=>'Iceland',
                'IT'=>'Italy',
                'JE'=>'Jersey',
                'JM'=>'Jamaica',
                'JO'=>'Jordan',
                'JP'=>'Japan',
                'KE'=>'Kenya',
                'KG'=>'Kyrgyzstan',
                'KH'=>'Cambodia',
                'KI'=>'Kiribati',
                'KM'=>'Comoros',
                'KN'=>'Saint Kitts and Nevis',
                'KP'=>'Korea (Democratic Peoples Republic of)',
                'KR'=>'Korea Republic of',
                'KW'=>'Kuwait',
                'KY'=>'Cayman Islands',
                'KZ'=>'Kazakhstan',
                'LA'=>'Lao Peoples Democratic Republic',
                'LB'=>'Lebanon',
                'LC'=>'Saint Lucia',
                'LI'=>'Liechtenstein',
                'LK'=>'Sri Lanka',
                'LR'=>'Liberia',
                'LS'=>'Lesotho',
                'LT'=>'Lithuania',
                'LU'=>'Luxembourg',
                'LV'=>'Latvia',
                'LY'=>'Libya',
                'MA'=>'Morocco',
                'MC'=>'Monaco',
                'MD'=>'Moldova Republic of',
                'ME'=>'Montenegro',
                'MF'=>'Saint Martin (French part)',
                'MG'=>'Madagascar',
                'MH'=>'Marshall Islands',
                'MK'=>'North Macedonia',
                'ML'=>'Mali',
                'MM'=>'Myanmar',
                'MN'=>'Mongolia',
                'MO'=>'Macao',
                'MP'=>'Northern Mariana Islands',
                'MQ'=>'Martinique',
                'MR'=>'Mauritania',
                'MS'=>'Montserrat',
                'MT'=>'Malta',
                'MU'=>'Mauritius',
                'MV'=>'Maldives',
                'MW'=>'Malawi',
                'MX'=>'Mexico',
                'MY'=>'Malaysia',
                'MZ'=>'Mozambique',
                'NA'=>'Namibia',
                'NC'=>'New Caledonia',
                'NE'=>'Niger',
                'NF'=>'Norfolk Island',
                'NG'=>'Nigeria',
                'NI'=>'Nicaragua',
                'NL'=>'Netherlands',
                'NO'=>'Norway',
                'NP'=>'Nepal',
                'NR'=>'Nauru',
                'NU'=>'Niue',
                'NZ'=>'New Zealand',
                'OM'=>'Oman',
                'PA'=>'Panama',
                'PE'=>'Peru',
                'PF'=>'French Polynesia',
                'PG'=>'Papua New Guinea',
                'PH'=>'Philippines',
                'PK'=>'Pakistan',
                'PL'=>'Poland',
                'PM'=>'Saint Pierre and Miquelon',
                'PN'=>'Pitcairn',
                'PR'=>'Puerto Rico',
                'PS'=>'Palestine State of',
                'PT'=>'Portugal',
                'PW'=>'Palau',
                'PY'=>'Paraguay',
                'QA'=>'Qatar',
                'RE'=>'Reunion',
                'RO'=>'Romania',
                'RS'=>'Serbia',
                'RU'=>'Russian Federation',
                'RW'=>'Rwanda',
                'SA'=>'Saudi Arabia',
                'SB'=>'Solomon Islands',
                'SC'=>'Seychelles',
                'SD'=>'Sudan',
                'SE'=>'Sweden',
                'SG'=>'Singapore',
                'SH'=>'Saint Helena Ascension and Tristan da Cunha',
                'SI'=>'Slovenia',
                'SJ'=>'Svalbard and Jan Mayen',
                'SK'=>'Slovakia',
                'SL'=>'Sierra Leone',
                'SM'=>'San Marino',
                'SN'=>'Senegal',
                'SO'=>'Somalia',
                'SR'=>'Suriname',
                'SS'=>'South Sudan',
                'ST'=>'Sao Tome and Principe',
                'SV'=>'El Salvador',
                'SX'=>'Sint Maarten (Dutch part)',
                'SY'=>'Syrian Arab Republic',
                'SZ'=>'Eswatini',
                'TC'=>'Turks and Caicos Islands',
                'TD'=>'Chad',
                'TF'=>'French Southern Territories',
                'TG'=>'Togo',
                'TH'=>'Thailand',
                'TJ'=>'Tajikistan',
                'TK'=>'Tokelau',
                'TL'=>'Timor-Leste',
                'TM'=>'Turkmenistan',
                'TN'=>'Tunisia',
                'TO'=>'Tonga',
                'TR'=>'Turkey',
                'TT'=>'Trinidad and Tobago',
                'TV'=>'Tuvalu',
                'TW'=>'Taiwan Province of China',
                'TZ'=>'Tanzania United Republic of',
                'UA'=>'Ukraine',
                'UG'=>'Uganda',
                'UM'=>'United States Minor Outlying Islands',
                'US'=>'United States of America',
                'UY'=>'Uruguay',
                'UZ'=>'Uzbekistan',
                'VA'=>'Holy See',
                'VC'=>'Saint Vincent and the Grenadines',
                'VE'=>'Venezuela (Bolivarian Republic of)',
                'VG'=>'Virgin Islands (British)',
                'VI'=>'Virgin Islands (U.S.)',
                'VN'=>'Viet Nam',
                'VU'=>'Vanuatu',
                'WF'=>'Wallis and Futuna',
                'WS'=>'Samoa',
                'YE'=>'Yemen',
                'YT'=>'Mayotte',
                'ZA'=>'South Africa',
                'ZM'=>'Zambia',
                'ZW'=>'Zimbabwe',
            ],
        ];
        if(isset($array[$key])){
            return $array[$key];
        }
        return [];
	}
} //end class
