<?php
/**
 * Plugin Name: WooCommerce Applepay Checkout Gateway v2
 * Plugin URI: https://www.totalprocessing.com/woocommerce/applepay
 * Description: Accept all major credit and debit cards enlisted to Applepay wallets. Fast, seamless, and flexible.
 * Version: 4.0
 * Author: Total Processing Limited
 * Author URI: https://www.totalprocessing.com
 * Copyright: 2021 Total Processing Limited.
 * License: GNU General Public License v3.0
 * License URI: http://www.gnu.org/licenses/gpl-3.0.html
 * Text Domain: woocommerce-gateway-totalprocessing-applepay-v2
 * Domain Path: /languages
 * WC tested up to: 4.9.1
 * WC requires at least: 4.0
 */
/**
 * Copyright (c) 2021 Total Processing Limited.
 *
 */
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}
function wc_gateway_tpapv2_activation() {
	if(get_option('wc_gateway_tpapv2_permalinks_flushed') === false){
		add_option('wc_gateway_tpapv2_permalinks_flushed',0);
	} else {
		update_option('wc_gateway_tpapv2_permalinks_flushed',0);
	}
}
register_activation_hook( __FILE__, 'wc_gateway_tpapv2_activation' );
function wc_gateway_tpapv2_deactivation() {
    delete_option('wc_gateway_tpapv2_permalinks_flushed');
	delete_option('wc_gateway_tpapv2_merchantIdentifier_v2');
    return false;
}
register_deactivation_hook( __FILE__, 'wc_gateway_tpapv2_deactivation' );
add_filter( 'woocommerce_payment_gateways', 'wc_gateway_tpapv2_register' );
function wc_gateway_tpapv2_register( $gateways ) {
	$gateways[] = 'WC_Gateway_TPAP2_Plugin';
	return $gateways;
}
//new section updater...
require __DIR__ . '/updater/plugin-update-checker.php';
$myUpdateChecker = Puc_v4_Factory::buildUpdateChecker(
	'https://plugins.totalprocessing.com/wp/tpap-v2/info.json',
	__FILE__,
	'woocommerce-gateway-totalprocessing-applepay-v2'
);
//end updater
add_action( 'init', 'tpapv2_custom_query_vars' );
function tpapv2_custom_query_vars() {
    global $wp;
    $wp->add_query_var( 'apple-developer-merchantid-domain-association' );
	$regex = '^\.well-known\/apple-developer-merchantid-domain-association';
    $redirect = 'index.php?apple-developer-merchantid-domain-association=1';
    add_rewrite_rule( $regex, $redirect, 'top' );
	if( !get_option('wc_gateway_tpapv2_permalinks_flushed') ) {
        flush_rewrite_rules(true);
        update_option('wc_gateway_tpapv2_permalinks_flushed', 1);
    }
}
add_action( 'plugins_loaded', 'wc_gateway_tpapv2_init' );
function wc_gateway_tpapv2_appleValidation( $wp ) {
    if ( isset( $wp->query_vars['apple-developer-merchantid-domain-association'] ) && '1' === $wp->query_vars['apple-developer-merchantid-domain-association'] ) {
        header( 'Content-Type: text/plain' );
		//header( 'Content-Type: application/octet-stream' );
        $path = dirname(__FILE__).'/appleValidation/apple-developer-merchantid-domain-association.txt';
		echo esc_html( file_get_contents( $path ) );
        exit;
    }
}
add_action( 'parse_request', 'wc_gateway_tpapv2_appleValidation', 10, 1 );
function wc_gateway_tpapv2_init() {
	require_once( 'includes/class-wc-gateway-tpapv2-plugin.php' );
	$tpapv2 = new WC_Gateway_TPAP2_Plugin();
	$tpapv2->run( plugin_basename( __FILE__ ) );
}